from unittest import TestCase

from pcs.common.reports import codes as report_codes
from pcs.lib.commands import stonith_agent as lib

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    start_tag_error_text,
)
from pcs_test.tools.command_env import get_env_tools

_fixture_fenced_xml = """
    <?xml version="1.0"?><!DOCTYPE resource-agent SYSTEM "ra-api-1.dtd">
    <resource-agent name="pacemaker-fenced">
        <version>1.0</version>
        <parameters>
            <parameter name="fenced-param">
                <shortdesc>testing fenced parameter</shortdesc>
                <longdesc>with a longdesc</longdesc>
            </parameter>
        </parameters>
    </resource-agent>
"""


def _fixture_parameter(name, shortdesc, longdesc=None):
    return {
        "advanced": False,
        "default": None,
        "deprecated": False,
        "deprecated_by": [],
        "deprecated_desc": None,
        "enum_values": None,
        "longdesc": longdesc,
        "name": name,
        "reloadable": False,
        "required": False,
        "shortdesc": shortdesc,
        "type": "string",
        "unique_group": None,
    }


_fixture_fenced_parsed = [
    _fixture_parameter(
        "fenced-param",
        "testing fenced parameter",
        "testing fenced parameter\nwith a longdesc",
    )
]


class ListAgents(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "stonith", "\n".join(["fence_apc", "fence_xvm", "fence_dummy"])
        )

    @staticmethod
    def _fixture_agent_struct(name):
        return {
            "name": f"stonith:{name}",
            "standard": "stonith",
            "provider": None,
            "type": name,
            "shortdesc": None,
            "longdesc": None,
            "parameters": [],
            "actions": [],
            "default_actions": [],
        }

    @staticmethod
    def _fixture_agent_metadata(name):
        return f"""
            <resource-agent name="{name}">
                <shortdesc>short {name}</shortdesc>
                <longdesc>long {name}</longdesc>
                <parameters>
                    <parameter name="own-param">
                        <shortdesc>testing own parameter</shortdesc>
                    </parameter>
                </parameters>
                <actions>
                </actions>
            </resource-agent>
            """

    def test_list_all(self):
        self.assertEqual(
            lib.list_agents(self.env_assist.get_env(), False, None),
            [
                self._fixture_agent_struct("fence_apc"),
                self._fixture_agent_struct("fence_dummy"),
                self._fixture_agent_struct("fence_xvm"),
            ],
        )

    def test_search(self):
        self.assertEqual(
            lib.list_agents(self.env_assist.get_env(), False, "M"),
            [
                self._fixture_agent_struct("fence_dummy"),
                self._fixture_agent_struct("fence_xvm"),
            ],
        )

    def test_describe(self):
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_apc",
            stdout=self._fixture_agent_metadata("stonith:fence_apc"),
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.fence_apc",
        )
        self.config.runner.pcmk.load_fake_agent_metadata(
            stdout=_fixture_fenced_xml
        )
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_dummy",
            agent_is_missing=True,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.fence_dummy",
        )
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_xvm",
            stdout=self._fixture_agent_metadata("stonith:fence_xvm"),
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.fence_xvm",
        )
        agent_stub = {
            "parameters": [
                _fixture_parameter("own-param", "testing own parameter")
            ]
            + _fixture_fenced_parsed,
            "actions": [],
            "default_actions": [
                {
                    "name": "monitor",
                    "interval": "60s",
                    "OCF_CHECK_LEVEL": None,
                    "automatic": False,
                    "on_target": False,
                    "role": None,
                    "start-delay": None,
                    "timeout": None,
                }
            ],
        }
        self.assertEqual(
            lib.list_agents(self.env_assist.get_env(), True, None),
            [
                dict(
                    name="stonith:fence_apc",
                    standard="stonith",
                    provider=None,
                    type="fence_apc",
                    shortdesc="short stonith:fence_apc",
                    longdesc="long stonith:fence_apc",
                    **agent_stub,
                ),
                dict(
                    name="stonith:fence_xvm",
                    standard="stonith",
                    provider=None,
                    type="fence_xvm",
                    shortdesc="short stonith:fence_xvm",
                    longdesc="long stonith:fence_xvm",
                    **agent_stub,
                ),
            ],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="stonith:fence_dummy",
                    reason=(
                        "Agent stonith:fence_dummy not found or does not support "
                        "meta-data: Invalid argument (22)\nMetadata query for "
                        "stonith:fence_dummy failed: Input/output error"
                    ),
                )
            ]
        )


class DescribeAgent(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_dummy",
            stdout="""
                <resource-agent name="fence_dummy">
                    <shortdesc>short desc</shortdesc>
                    <longdesc>long desc</longdesc>
                    <parameters>
                        <parameter name="own-param">
                            <shortdesc>testing own parameter</shortdesc>
                        </parameter>
                    </parameters>
                    <actions>
                    </actions>
                </resource-agent>
            """,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.fence_apc",
        )
        self.config.runner.pcmk.load_fake_agent_metadata(
            stdout=_fixture_fenced_xml
        )

        self.assertEqual(
            lib.describe_agent(self.env_assist.get_env(), "fence_dummy"),
            {
                "name": "stonith:fence_dummy",
                "standard": "stonith",
                "provider": None,
                "type": "fence_dummy",
                "shortdesc": "short desc",
                "longdesc": "long desc",
                "parameters": [
                    _fixture_parameter("own-param", "testing own parameter")
                ]
                + _fixture_fenced_parsed,
                "actions": [],
                "default_actions": [
                    {
                        "name": "monitor",
                        "interval": "60s",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "timeout": None,
                    }
                ],
            },
        )

    def test_fail(self):
        self.config.runner.pcmk.load_agent(
            agent_name="stonith:fence_dummy",
            stdout="this is not a proper metadata xml",
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )

        assert_raise_library_error(
            lambda: lib.describe_agent(self.env_assist.get_env(), "fence_dummy")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="stonith:fence_dummy",
                    reason=start_tag_error_text(),
                ),
            ]
        )

    def test_invalid_name(self):
        assert_raise_library_error(
            lambda: lib.describe_agent(
                self.env_assist.get_env(), "stonith:fence_dummy"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_STONITH_AGENT_NAME,
                    name="stonith:fence_dummy",
                )
            ]
        )



from typing import NewType

FileTypeCode = NewType("FileTypeCode", str)

BOOTH_CONFIG = FileTypeCode("BOOTH_CONFIG")
BOOTH_KEY = FileTypeCode("BOOTH_KEY")
CIB = FileTypeCode("CIB")
COROSYNC_AUTHKEY = FileTypeCode("COROSYNC_AUTHKEY")
COROSYNC_CONF = FileTypeCode("COROSYNC_CONF")
COROSYNC_QDEVICE_NSSDB = FileTypeCode("COROSYNC_QDEVICE_NSSDB")
COROSYNC_QNETD_NSSDB = FileTypeCode("COROSYNC_QNETD_NSSDB")
PACEMAKER_AUTHKEY = FileTypeCode("PACEMAKER_AUTHKEY")
PCSD_ENVIRONMENT_CONFIG = FileTypeCode("PCSD_ENVIRONMENT_CONFIG")
PCSD_SSL_CERT = FileTypeCode("PCSD_SSL_CERT")
PCSD_SSL_KEY = FileTypeCode("PCSD_SSL_KEY")
PCS_KNOWN_HOSTS = FileTypeCode("PCS_KNOWN_HOSTS")
PCS_SETTINGS_CONF = FileTypeCode("PCS_SETTINGS_CONF")
PCS_DR_CONFIG = FileTypeCode("PCS_DR_CONFIG")
PCS_USERS_CONF = FileTypeCode("PCS_USERS_CONF")



import datetime
import os
import signal
from asyncio import Event
from dataclasses import dataclass
from typing import (
    Any,
    Awaitable,
    List,
    Optional,
)

from pcs import settings
from pcs.common.async_tasks.dto import TaskResultDto
from pcs.common.async_tasks.types import (
    TaskFinishType,
    TaskKillReason,
    TaskState,
)
from pcs.common.interface.dto import ImplementsToDto
from pcs.common.reports.dto import ReportItemDto
from pcs.lib.auth.types import AuthUser

from .types import Command
from .worker.types import (
    Message,
    TaskExecuted,
    TaskFinished,
    WorkerCommand,
)


class UnknownMessageError(Exception):
    """
    A message with an unknown message type was received
    """

    def __init__(self, unknown_message: Message):
        super().__init__(self)
        self.payload_type = type(unknown_message.payload).__name__


_STATE_CHANGES_SEQUENCE = (
    TaskState.CREATED,
    TaskState.QUEUED,
    TaskState.EXECUTED,
    TaskState.FINISHED,
)


@dataclass(frozen=True)
class TaskConfig:
    abandoned_timeout: int = settings.task_abandoned_timeout_seconds
    unresponsive_timeout: int = settings.task_unresponsive_timeout_seconds
    deletion_timeout: int = settings.task_deletion_timeout_seconds


class Task(ImplementsToDto):
    """
    Task's representation in the scheduler
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        task_ident: str,
        command: Command,
        auth_user: AuthUser,
        config: TaskConfig,
    ) -> None:
        self._config = config
        self._task_ident: str = task_ident
        self._command: Command = command
        self._auth_user = auth_user
        self._reports: List[ReportItemDto] = []
        self._result: Any = None
        self._state: TaskState = TaskState.CREATED
        self._task_finish_type: TaskFinishType = TaskFinishType.UNFINISHED
        self._kill_reason: Optional[TaskKillReason] = None
        self._last_message_at: Optional[datetime.datetime] = None
        self._execution_started_at: Optional[datetime.datetime] = None
        self._worker_pid: int = -1
        self._finished_event = Event()
        self._to_delete_timestamp: Optional[datetime.datetime] = None

    @property
    def state(self) -> TaskState:
        return self._state

    @state.setter
    def state(self, state: TaskState) -> None:
        """
        Sets a new task state
        :param state: New task state
        """
        self._set_state(state)

    def _set_state(self, state: TaskState) -> None:
        try:
            if _STATE_CHANGES_SEQUENCE.index(
                self._state
            ) >= _STATE_CHANGES_SEQUENCE.index(state):
                raise AssertionError(
                    f"Invalid Task state change: {self._state} => {state}"
                )
        except ValueError as e:
            raise AssertionError(f"Invalid Task state: {state}") from e
        self._state = state
        if self.state == TaskState.FINISHED:
            self._finished_event.set()
        elif self.state == TaskState.EXECUTED:
            self._execution_started_at = datetime.datetime.now()

    @property
    def task_ident(self) -> str:
        return self._task_ident

    @property
    def auth_user(self) -> AuthUser:
        return self._auth_user

    def wait_until_finished(self) -> Awaitable[Any]:
        return self._finished_event.wait()

    def _get_last_updated_timestamp(self) -> Optional[datetime.datetime]:
        """
        Helper function for getting timestamp of the last message received

        This function forces a time of setting the task to FINISHED for the
        task that haven't received any messages.
        :return: Date and time of receiving the last message
        """
        if self._last_message_at is None and self.state == TaskState.FINISHED:
            # Update the timestamp of last message
            self._task_updated()
        return self._last_message_at

    def _is_timed_out(self, timeout_s: int) -> bool:
        """
        Helper function for watching time since last message was delivered
        :param timeout_s: Timeout in seconds
        :return: True if last message was delivered sooner than now - timeout,
            False otherwise
        """
        last_message_at = self._get_last_updated_timestamp()
        if last_message_at:
            return (
                datetime.datetime.now() - last_message_at
                > datetime.timedelta(seconds=timeout_s)
            )
        return False

    def is_abandoned(self) -> bool:
        """
        Checks that the task information were not queried after its finish
        :return: True if task was not queried for information before timeout,
            False otherwise
        """
        if self.state == TaskState.FINISHED:
            # Last message of finished task is notification of its completion
            # and thus marks the time of its completion
            return self._is_timed_out(self._config.abandoned_timeout)
        # Even if the client dies before task finishes, task is going
        # to finish and will be garbage collected as abandoned
        return False

    def is_defunct(self, timeout: Optional[int] = None) -> bool:
        """
        Checks that the task is not behaving as expected

        Healthy tasks are expected to send reports or other messages. If the
        task becomes stuck or malfunctions in some way, it's probable that it
        will stop sending reports or messages.
        :return: True if no messages were received during timeout period since
            the last message, False otherwise
        """
        if timeout is None or timeout < 0:
            timeout = self._config.unresponsive_timeout
        if self.state == TaskState.EXECUTED:
            return self._is_timed_out(timeout)
        return False

    def _task_updated(self) -> None:
        """
        Helper function for setting the last message timestamp to now
        """
        self._last_message_at = datetime.datetime.now()

    def is_kill_requested(self) -> bool:
        """
        Reports if the task needs to be killed

        :return: True for tasks marked for killing
        """
        return self._kill_reason is not None

    def request_kill(self, reason: TaskKillReason) -> None:
        """
        Marks the task for cleanup to the garbage collector
        :param reason: Reason for killing the task
        """
        self._kill_reason = reason

    def request_deletion(self) -> None:
        """
        Marks the task for deletion. .is_deletion_requested() method will
        return True after TaskConfig.deletion_timeout timeout.
        """
        if self._to_delete_timestamp is None:
            self._to_delete_timestamp = (
                datetime.datetime.now()
                + datetime.timedelta(seconds=self._config.deletion_timeout)
            )

    def is_deletion_requested(self) -> bool:
        """
        Indicates whether a task should be deleted.
        """
        if self._to_delete_timestamp is None:
            return False
        return datetime.datetime.now() >= self._to_delete_timestamp

    def kill(self) -> None:
        """
        Terminates the task and/or changes its state

        CREATED tasks are already prevented from being scheduled by requesting
        to kill them, only their state gets corrected here.
        EXECUTED tasks are terminated by by sending SIGTERM to their worker
        process and their state is changed here.
        """
        if self.state in (
            TaskState.QUEUED,
            TaskState.FINISHED,
        ):
            return
        if self.state == TaskState.EXECUTED:
            try:
                os.kill(self._worker_pid, signal.SIGTERM)
            except ProcessLookupError:
                # PID doesn't exist, process might have died on its own or
                # finished even in the time since task state was checked. Since
                # the killing wasn't successful, don't change the state
                return

        self._set_state(TaskState.FINISHED)
        self._task_finish_type = TaskFinishType.KILL

    # Message handlers
    def receive_message(self, message: Message) -> None:
        """
        Main message handler
        :param message: Message instance
        """
        if isinstance(message.payload, ReportItemDto):
            self._store_reports(message.payload)
        elif isinstance(message.payload, TaskExecuted):
            self._message_executed(message.payload)
        elif isinstance(message.payload, TaskFinished):
            self._message_finished(message.payload)
        else:
            raise UnknownMessageError(message)
        self._task_updated()

    def _message_executed(self, message_payload: TaskExecuted) -> None:
        """
        Handler for scheduler's TaskExecuted messages
        """
        self._worker_pid = message_payload.worker_pid
        self._set_state(TaskState.EXECUTED)

    def _message_finished(self, message_payload: TaskFinished) -> None:
        """
        Handler for scheduler's TaskFinished messages
        """
        self._result = message_payload.result
        self._set_state(TaskState.FINISHED)
        self._task_finish_type = message_payload.task_finish_type
        os.kill(self._worker_pid, signal.SIGCONT)

    def _store_reports(self, message_payload: ReportItemDto) -> None:
        """
        Handler for PCS reports
        """
        self._reports.append(message_payload)

    # Type conversions
    def to_worker_command(self) -> WorkerCommand:
        """
        Creates structure for sending task to a worker process
        :return: Instance with task identifier, command and parameters
        """
        return WorkerCommand(self._task_ident, self._command, self._auth_user)

    def to_dto(self) -> TaskResultDto:
        """
        Prepares response for task information query
        :return: DTO object with information about the task
        """
        return TaskResultDto(
            self._task_ident,
            self._command.command_dto,
            self._reports,
            self.state,
            self._task_finish_type,
            self._kill_reason,
            self._result,
        )



from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common.const import (
    PcmkOnFailAction,
    PcmkRoleType,
)
from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.nvset import CibNvsetDto

OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME = "OCF_CHECK_LEVEL"


@dataclass(frozen=True)
class CibResourceOperationDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    id: str  # pylint: disable=invalid-name
    name: str
    interval: str
    description: Optional[str]
    # exactly one of start_delay and interval_origin should be defined
    start_delay: Optional[str]
    interval_origin: Optional[str]
    timeout: Optional[str]
    enabled: Optional[bool]
    record_pending: Optional[bool]
    role: Optional[PcmkRoleType]
    on_fail: Optional[PcmkOnFailAction]
    meta_attributes: Sequence[CibNvsetDto]
    instance_attributes: Sequence[CibNvsetDto]


@dataclass(frozen=True)
class ListCibResourceOperationDto(DataTransferObject):
    operations: Sequence[CibResourceOperationDto]



import re

TRANSPORT_DEFAULT = "knet"
TRANSPORTS_KNET = ("knet",)
TRANSPORTS_UDP = ("udp", "udpu")
TRANSPORTS_ALL = TRANSPORTS_KNET + TRANSPORTS_UDP

TRANSPORT_KNET_CRYPTO_OPTIONS = (
    "cipher",
    "hash",
    "model",
)

TRANSPORT_KNET_COMPRESSION_OPTIONS = (
    "level",
    "model",
    "threshold",
)

TRANSPORT_KNET_GENERIC_OPTIONS = (
    "ip_version",  # It tells knet which IP to prefer.
    "knet_pmtud_interval",
    "link_mode",
)

TRANSPORT_UDP_GENERIC_OPTIONS = (
    "ip_version",
    "netmtu",
)

TRANSPORT_RUNTIME_CHANGE_BANNED_OPTIONS = (
    "ip_version",
    "netmtu",
)

TOTEM_OPTIONS = (
    "block_unlisted_ips",
    "consensus",
    "downcheck",
    "fail_recv_const",
    "heartbeat_failures_allowed",
    "hold",
    "join",
    "max_messages",
    "max_network_delay",
    "merge",
    "miss_count_const",
    "send_join",
    "seqno_unchanged_const",
    "token",
    "token_coefficient",
    "token_retransmit",
    "token_retransmits_before_loss_const",
    "window_size",
)

LINKS_UDP_MIN = 1
LINKS_UDP_MAX = 1
LINKS_KNET_MIN = 1
LINKS_KNET_MAX = 8
LINKS_MAX = max(LINKS_KNET_MAX, LINKS_UDP_MAX)

NODE_OPTIONS = ["name", "nodeid"] + [f"ring{i}_addr" for i in range(LINKS_MAX)]

IP_VERSION_4 = "ipv4"
IP_VERSION_6 = "ipv6"
IP_VERSION_46 = "ipv4-6"
IP_VERSION_64 = "ipv6-4"
IP_VERSION_VALUES = (IP_VERSION_4, IP_VERSION_6, IP_VERSION_46, IP_VERSION_64)

# This must list ALL knet options even if they translate to themselves.
# The list is used to get a set of allowed options.
LINK_OPTIONS_KNET_TRANSLATION = (
    # (user facing name, corosync.conf name)
    ("linknumber", "linknumber"),
    ("link_priority", "knet_link_priority"),
    ("mcastport", "mcastport"),
    ("ping_interval", "knet_ping_interval"),
    ("ping_precision", "knet_ping_precision"),
    ("ping_timeout", "knet_ping_timeout"),
    ("pong_count", "knet_pong_count"),
    ("transport", "knet_transport"),
)
# This is what pcs expects to get from users. Actual names in corosync.conf are
# different.
LINK_OPTIONS_KNET_USER = tuple(
    sorted([pair[0] for pair in LINK_OPTIONS_KNET_TRANSLATION])
)
# This is what corsync works with.
LINK_OPTIONS_KNET_COROSYNC = tuple(
    sorted([pair[1] for pair in LINK_OPTIONS_KNET_TRANSLATION])
)

LINK_OPTIONS_UDP = (
    "bindnetaddr",
    "broadcast",
    "mcastaddr",
    "mcastport",
    "ttl",
)

QUORUM_OPTIONS = (
    "auto_tie_breaker",
    "last_man_standing",
    "last_man_standing_window",
    "wait_for_all",
)

QUORUM_OPTIONS_INCOMPATIBLE_WITH_QDEVICE = (
    "auto_tie_breaker",
    "last_man_standing",
    "last_man_standing_window",
)

OPTION_NAME_RE = re.compile(r"^[-_/a-zA-Z0-9]+$")
QUORUM_DEVICE_HEURISTICS_EXEC_NAME_RE = re.compile(r"^exec_[-_/a-zA-Z0-9]+$")



from dataclasses import dataclass
from typing import (
    Any,
    Union,
)

from pcs.common.async_tasks.types import TaskFinishType
from pcs.common.reports import ReportItemDto
from pcs.daemon.async_tasks.types import Command
from pcs.lib.auth.types import AuthUser


@dataclass(frozen=True)
class TaskExecuted:
    worker_pid: int


@dataclass(frozen=True)
class TaskFinished:
    task_finish_type: TaskFinishType
    result: Any


@dataclass(frozen=True)
class Message:
    task_ident: str
    payload: Union[
        ReportItemDto,
        TaskExecuted,
        TaskFinished,
    ]


@dataclass(frozen=True)
class WorkerCommand:
    task_ident: str
    command: Command
    auth_user: AuthUser



from pcs.daemon import (
    ruby_pcsd,
    session,
)
from pcs.daemon.app.auth import (
    NotAuthorizedException,
    SessionAuthProvider,
)
from pcs.daemon.app.common import (
    LegacyApiHandler,
    RoutesType,
)
from pcs.daemon.app.sinatra_common import SinatraMixin
from pcs.daemon.app.ui_common import AjaxMixin
from pcs.lib.auth.provider import AuthProvider


class SinatraAjaxProtected(LegacyApiHandler, SinatraMixin, AjaxMixin):
    """
    SinatraAjaxProtected handles urls that calls the ajax Sinatra GUI functions.
    It allows to use this urls only for ajax calls.
    """

    _auth_provider: SessionAuthProvider

    def initialize(
        self,
        session_storage: session.Storage,
        ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
        auth_provider: AuthProvider,
    ) -> None:
        # pylint: disable=arguments-differ
        self.initialize_sinatra(ruby_pcsd_wrapper)
        self._auth_provider = SessionAuthProvider(
            self, auth_provider, session_storage
        )

    def prepare(self) -> None:
        if not self.is_ajax:
            raise self.unauthorized()
        self._auth_provider.init_session()

    async def _handle_request(self):
        try:
            auth_user = await self._auth_provider.auth_by_sid()
        except NotAuthorizedException as e:
            raise self.unauthorized() from e
        self._auth_provider.update_session(auth_user)
        if self._auth_provider.is_sid_in_request_cookies():
            self._auth_provider.put_request_cookies_sid_to_response_cookies_sid()

        result = await self.ruby_pcsd_wrapper.request(
            auth_user,
            self.request,
        )
        self.send_sinatra_result(result)


def get_routes(
    session_storage: session.Storage,
    auth_provider: AuthProvider,
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
) -> RoutesType:
    return [
        # The protection by session was moved from ruby code to python code
        # (tornado).
        (
            r"/.*",
            SinatraAjaxProtected,
            dict(
                session_storage=session_storage,
                ruby_pcsd_wrapper=ruby_pcsd_wrapper,
                auth_provider=auth_provider,
            ),
        ),
    ]



from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Any,
    List,
    Mapping,
    Optional,
    Union,
)
from unittest import TestCase

from pcs.common.reports import (
    dto,
    item,
    types,
)

REPORT_CODE = types.MessageCode("REPORT_CODE")
OUTER_REPORT_CODE = types.MessageCode("OUTER_REPORT_CODE")


@dataclass(frozen=True)
class ReportMessage(item.ReportItemMessage):
    string: str
    union_str_or_list_of_str: Union[str, List[str]]
    list_of_str: List[str] = field(default_factory=list)
    optional_list_of_int: Optional[List[int]] = None
    optional_mapping_str_to_any: Optional[Mapping[str, Any]] = None
    _code = REPORT_CODE

    @property
    def message(self):
        return "a message"


@dataclass(frozen=True)
class OuterReportMessage(item.ReportItemMessage):
    optional_inner_msg: Optional[ReportMessage]
    optional_string: Optional[str] = None
    _code = OUTER_REPORT_CODE

    @property
    def message(self):
        return "an outer message"


class ReportItemMessageToDtoTest(TestCase):
    def test_all_specified(self):
        self.assertEqual(
            ReportMessage(
                "a string",
                "another string",
                ["str1", "str0"],
                [0, 1, 0],
                {"str1": 1, "key": "val", "another": False},
            ).to_dto(),
            dto.ReportItemMessageDto(
                REPORT_CODE,
                "a message",
                dict(
                    string="a string",
                    union_str_or_list_of_str="another string",
                    list_of_str=["str1", "str0"],
                    optional_list_of_int=[0, 1, 0],
                    optional_mapping_str_to_any=dict(
                        str1=1,
                        key="val",
                        another=False,
                    ),
                ),
            ),
        )

    def test_minimal(self):
        self.assertEqual(
            ReportMessage("a string", ["str1", "str0"]).to_dto(),
            dto.ReportItemMessageDto(
                REPORT_CODE,
                "a message",
                dict(
                    string="a string",
                    union_str_or_list_of_str=["str1", "str0"],
                    list_of_str=[],
                    optional_list_of_int=None,
                    optional_mapping_str_to_any=None,
                ),
            ),
        )

    def test_alternatives(self):
        self.assertEqual(
            ReportMessage(
                "a string",
                ["str1", "str0"],
                ["str1", "str0", "str3"],
                optional_list_of_int=None,
                optional_mapping_str_to_any={},
            ).to_dto(),
            dto.ReportItemMessageDto(
                REPORT_CODE,
                "a message",
                dict(
                    string="a string",
                    union_str_or_list_of_str=["str1", "str0"],
                    list_of_str=["str1", "str0", "str3"],
                    optional_list_of_int=None,
                    optional_mapping_str_to_any={},
                ),
            ),
        )

    def test_with_inner_msg(self):
        self.assertEqual(
            OuterReportMessage(
                ReportMessage("a string", ["str1", "str0"]), "a str"
            ).to_dto(),
            dto.ReportItemMessageDto(
                OUTER_REPORT_CODE,
                "an outer message",
                dict(
                    optional_inner_msg=dto.ReportItemMessageDto(
                        REPORT_CODE,
                        "a message",
                        dict(
                            string="a string",
                            union_str_or_list_of_str=["str1", "str0"],
                            list_of_str=[],
                            optional_list_of_int=None,
                            optional_mapping_str_to_any=None,
                        ),
                    ),
                    optional_string="a str",
                ),
            ),
        )



from typing import Optional

from pcs.common.services_dto import (
    ServicesInfoResultDto,
    ServiceStatusDto,
)
from pcs.common.types import StringIterable
from pcs.lib.env import LibraryEnvironment


def start_service(
    env: LibraryEnvironment,
    service: str,
    instance: Optional[str],
) -> None:
    env.service_manager.start(service, instance)


def stop_service(
    env: LibraryEnvironment,
    service: str,
    instance: Optional[str],
) -> None:
    env.service_manager.stop(service, instance)


def enable_service(
    env: LibraryEnvironment,
    service: str,
    instance: Optional[str],
) -> None:
    env.service_manager.enable(service, instance)


def disable_service(
    env: LibraryEnvironment,
    service: str,
    instance: Optional[str],
) -> None:
    env.service_manager.disable(service, instance)


def get_services_info(
    env: LibraryEnvironment,
    services: StringIterable,
    installed: bool = False,
    enabled: bool = False,
    running: bool = False,
) -> ServicesInfoResultDto:
    """
    services -- status of these services will be returned
    installed -- check whether specified services are installed
    enabled -- check whether specified services are enabled
    running -- check whether specified services are running

    Returns status of specified services based on requested information
    (`installed`, `running`, `enabled`). If `installed` is `True`, all
    installed services will be returned, but enabled nor running is not checked
    for services not specified in `services`
    """
    service_set = set(services)
    return ServicesInfoResultDto(
        [
            ServiceStatusDto(
                service,
                (
                    env.service_manager.is_installed(service)
                    if installed
                    else None
                ),
                (
                    env.service_manager.is_enabled(service)
                    if enabled and service in service_set
                    else None
                ),
                (
                    env.service_manager.is_running(service)
                    if running and service in service_set
                    else None
                ),
            )
            for service in service_set
            | set(
                env.service_manager.get_available_services()
                if installed
                else []
            )
        ]
    )



"""
This is the set of tools for testing commands (pcs.lib.commands).

The principle is to patch some parts of the library environment object
(pcs.lib.env.LibraryEnvironment) that is passed as first argument to each
command.

Important parts:
================
CallListBuilder + (Call)Queue
-----------------------------
Both objects stores list of calls (messages to the mocked parts of environment).
CallListBuilder is used in configuration phase (before command run) to build
the call list.
Queue is used in run phase (during command run) to check that everything is done
as expected.

Mocks (Runner, push_cib, ...)
-----------------------------
Mocks replaces real environment parts. Every Mock has an access to Queue.
Everytime when the mock obtain a message from tested command it takes expected
message from Queue. Then the mock compares expected and real message. When the
messages match each other then the Mock returns expected result. Otherwise Mock
call fails.

With each Mock comes Call that represent message appropriate for the concrete
Mock.

Config (with RunnerConfig, CibShortcuts, EnvConfig, ...)
--------------------------------------------------------
The tests use the Config for building list of expected calls (messages to
the mocked parts). Config stores list of calls in CallListBuilder.

EnvAssistant
-----------
EnvAssistant provides CallListBuilder to Config. When test requests an
environment (from the EnvAssistant) then the EnvAssistant:
* takes calls from Config and prepares the Queue (of calls)
* creates appropriate mock and provide them the Queue
* patches environment by appropriate mocks
* returns patched environment

When the test is done the EnvAssistant unpatches the environment and do requeired
checks (that whole Queue is consumed, that there was no extra reports, ...)

Example:
========
from unittest import TestCase

from pcs.lib.commands import resource
from pcs_test.tools.command_env import get_env_tools

class ExampleTest(TestCase):
    def test_success(self):
        env_assist, config = get_env_tools(test_case=self)
        (config
            .runner.cib.load()
            .runner.cib.push(
                resources='''
                    <resources>
                        <bundle id="B1">
                            <docker image="pcs:test" />
                        </bundle>
                    </resources>
                '''
            )
        )
        resource.bundle_create(
            self.env_assist.get_env(),
            "B1",
            "docker",
            container_options={"image": "pcs:test"},
            ensure_disabled=disabled,
            wait=False,
        )
"""
from pcs_test.tools.command_env.assistant import EnvAssistant
from pcs_test.tools.command_env.config import Config
from pcs_test.tools.command_env.tools import get_env_tools



from unittest import TestCase

from pcs.lib.permissions.config.facade import FacadeV2
from pcs.lib.permissions.config.types import (
    ClusterPermissions,
    ConfigV2,
    PermissionAccessType,
    PermissionEntry,
    PermissionTargetType,
)

_USER1 = PermissionEntry(
    name="user1",
    type=PermissionTargetType.USER,
    allow=[PermissionAccessType.READ, PermissionAccessType.GRANT],
)
_USER2 = PermissionEntry(
    name="user2",
    type=PermissionTargetType.USER,
    allow=[PermissionAccessType.SUPERUSER],
)
_GROUP1 = PermissionEntry(
    name="group1",
    type=PermissionTargetType.GROUP,
    allow=[
        PermissionAccessType.READ,
        PermissionAccessType.WRITE,
        PermissionAccessType.GRANT,
    ],
)
_GROUP2 = PermissionEntry(
    name="group2",
    type=PermissionTargetType.GROUP,
    allow=[PermissionAccessType.WRITE],
)
_CONFIG = ConfigV2(
    data_version=1,
    clusters=[],
    permissions=ClusterPermissions(
        local_cluster=[_USER1, _USER2, _GROUP1, _GROUP2]
    ),
)


class FacadeV2GetEntryTest(TestCase):
    def setUp(self):
        self.facade = FacadeV2(_CONFIG)

    def test_user_success(self):
        self.assertEqual(
            _USER1, self.facade.get_entry(_USER1.name, _USER1.type)
        )

    def test_user_non_existing(self):
        self.assertIsNone(
            self.facade.get_entry("non-existing", PermissionTargetType.USER)
        )

    def test_user_wrong_type(self):
        self.assertIsNone(
            self.facade.get_entry(_GROUP1.name, PermissionTargetType.USER)
        )

    def test_group_success(self):
        self.assertEqual(
            _GROUP2, self.facade.get_entry(_GROUP2.name, _GROUP2.type)
        )

    def test_group_non_existing(self):
        self.assertIsNone(
            self.facade.get_entry("non-existing", PermissionTargetType.GROUP)
        )

    def test_group_wrong_type(self):
        self.assertIsNone(
            self.facade.get_entry(_USER2.name, PermissionTargetType.GROUP)
        )



import base64
import os.path
from functools import partial
from typing import (
    Optional,
    cast,
)

from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.file import (
    FileAlreadyExists,
    RawFileError,
)
from pcs.common.reports import ReportProcessor
from pcs.common.reports import codes as report_codes
from pcs.common.reports.item import (
    ReportItem,
    get_severity,
)
from pcs.common.services.errors import ManageServiceError
from pcs.common.str_tools import join_multilines
from pcs.lib import (
    tools,
    validate,
)
from pcs.lib.booth import (
    config_files,
    config_validators,
    constants,
    resource,
    status,
)
from pcs.lib.cib.resource import (
    group,
    hierarchy,
    primitive,
)
from pcs.lib.cib.tools import (
    IdProvider,
    get_resources,
)
from pcs.lib.communication.booth import (
    BoothGetConfig,
    BoothSendConfig,
)
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import (
    GhostFile,
    RealFile,
    raw_file_error_report,
)
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.resource_agent import (
    ResourceAgentError,
    ResourceAgentFacade,
    ResourceAgentFacadeFactory,
    ResourceAgentName,
    UnableToGetAgentMetadata,
    resource_agent_error_to_report_item,
)
from pcs.lib.services import (
    ensure_is_systemd,
    is_systemd,
    service_exception_to_report,
)


def config_setup(
    env: LibraryEnvironment,
    site_list,
    arbitrator_list,
    instance_name=None,
    overwrite_existing=False,
):
    """
    create booth configuration

    env
    list site_list -- site addresses of multisite
    list arbitrator_list -- arbitrator addresses of multisite
    string instance_name -- booth instance name
    bool overwrite_existing -- allow overwriting existing files
    """
    instance_name = instance_name or constants.DEFAULT_INSTANCE_NAME
    report_processor = env.report_processor

    report_processor.report_list(
        config_validators.check_instance_name(instance_name)
    )
    report_processor.report_list(
        config_validators.create(site_list, arbitrator_list)
    )
    if report_processor.has_errors:
        raise LibraryError()

    booth_env = env.get_booth_env(instance_name)

    booth_conf = booth_env.create_facade(site_list, arbitrator_list)
    booth_conf.set_authfile(booth_env.key_path)

    conf_dir = (
        None
        if booth_env.ghost_file_codes
        else os.path.dirname(booth_env.config_path)
    )

    try:
        booth_env.key.write_raw(
            tools.generate_binary_key(
                random_bytes_count=settings.booth_authkey_bytes
            ),
            can_overwrite=overwrite_existing,
        )
        booth_env.config.write_facade(
            booth_conf, can_overwrite=overwrite_existing
        )
    except FileAlreadyExists as e:
        report_processor.report(
            ReportItem(
                severity=reports.item.get_severity(
                    reports.codes.FORCE,
                    overwrite_existing,
                ),
                message=reports.messages.FileAlreadyExists(
                    e.metadata.file_type_code,
                    e.metadata.path,
                ),
            )
        )
    except RawFileError as e:
        if conf_dir and not os.path.exists(conf_dir):
            report_processor.report(
                ReportItem.error(reports.messages.BoothPathNotExists(conf_dir))
            )
        else:
            report_processor.report(raw_file_error_report(e))
    if report_processor.has_errors:
        raise LibraryError()


def config_destroy(
    env: LibraryEnvironment,
    instance_name: Optional[str] = None,
    ignore_config_load_problems: bool = False,
) -> None:
    # pylint: disable=too-many-branches
    """
    remove booth configuration files

    env
    instance_name -- booth instance name
    ignore_config_load_problems -- delete as much as possible when unable to
        read booth configs for the given booth instance
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    found_instance_name = booth_env.instance_name
    _ensure_live_env(env, booth_env)

    booth_resource_list = resource.find_for_config(
        get_resources(env.get_cib()),
        booth_env.config_path,
    )
    if booth_resource_list:
        report_processor.report(
            ReportItem.error(
                reports.messages.BoothConfigIsUsed(
                    found_instance_name,
                    reports.const.BOOTH_CONFIG_USED_IN_CLUSTER_RESOURCE,
                    resource_name=str(booth_resource_list[0].get("id", "")),
                )
            )
        )
    # Only systemd is currently supported. Initd does not supports multiple
    # instances (here specified by name)
    if is_systemd(env.service_manager):
        if env.service_manager.is_running("booth", found_instance_name):
            report_processor.report(
                ReportItem.error(
                    reports.messages.BoothConfigIsUsed(
                        found_instance_name,
                        reports.const.BOOTH_CONFIG_USED_RUNNING_IN_SYSTEMD,
                    )
                )
            )

        if env.service_manager.is_enabled("booth", found_instance_name):
            report_processor.report(
                ReportItem.error(
                    reports.messages.BoothConfigIsUsed(
                        found_instance_name,
                        reports.const.BOOTH_CONFIG_USED_ENABLED_IN_SYSTEMD,
                    )
                )
            )
    if report_processor.has_errors:
        raise LibraryError()

    authfile_path = None
    try:
        booth_conf = booth_env.config.read_to_facade()
        authfile_path = booth_conf.get_authfile()
    except RawFileError as e:
        report_processor.report(
            raw_file_error_report(
                e,
                force_code=report_codes.FORCE,
                is_forced_or_warning=ignore_config_load_problems,
            )
        )
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(
                e,
                force_code=report_codes.FORCE,
                is_forced_or_warning=ignore_config_load_problems,
            )
        )
    if report_processor.has_errors:
        raise LibraryError()

    if authfile_path:
        authfile_dir, authfile_name = os.path.split(authfile_path)
        if (authfile_dir == settings.booth_config_dir) and authfile_name:
            try:
                key_file = FileInstance.for_booth_key(
                    authfile_name, ghost_file=False
                )
                # We're sure we have a RealFile instance as for_booth_key was
                # called with ghost_file=False
                cast(RealFile, key_file.raw_file).remove(
                    fail_if_file_not_found=False
                )
            except RawFileError as e:
                report_processor.report(
                    raw_file_error_report(
                        e,
                        force_code=report_codes.FORCE,
                        is_forced_or_warning=ignore_config_load_problems,
                    )
                )
        else:
            report_processor.report(
                ReportItem.warning(
                    reports.messages.BoothUnsupportedFileLocation(
                        authfile_path,
                        settings.booth_config_dir,
                        file_type_codes.BOOTH_KEY,
                    )
                )
            )
    if report_processor.has_errors:
        raise LibraryError()

    try:
        booth_env.config.raw_file.remove()
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))

    if report_processor.has_errors:
        raise LibraryError()


# TODO: remove once settings booth_enable_autfile_(set|unset)_enabled are removed
def _config_set_enable_authfile(
    env: LibraryEnvironment, value: bool, instance_name=None
) -> None:
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    if (value and not settings.booth_enable_authfile_set_enabled) or (
        not value and not settings.booth_enable_authfile_unset_enabled
    ):
        raise AssertionError()
    try:
        booth_conf = booth_env.config.read_to_facade()
        booth_conf.set_option(
            constants.AUTHFILE_FIX_OPTION, "yes" if value else ""
        )
        booth_env.config.write_facade(booth_conf, can_overwrite=True)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()


def config_set_enable_authfile(
    env: LibraryEnvironment, instance_name=None
) -> None:
    _config_set_enable_authfile(env, True, instance_name=instance_name)


def config_unset_enable_authfile(
    env: LibraryEnvironment, instance_name=None
) -> None:
    _config_set_enable_authfile(env, False, instance_name=instance_name)


def config_text(env: LibraryEnvironment, instance_name=None, node_name=None):
    """
    get configuration in raw format

    env
    string instance_name -- booth instance name
    string node_name -- get the config from specified node or local host if None
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    instance_name = booth_env.instance_name
    # It does not make any sense for the cli to read a ghost file and send it
    # to lib so that the lib could return it unchanged to cli. Just use 'cat'.
    # When node_name is specified, using ghost files doesn't make any sense
    # either.
    _ensure_live_env(env, booth_env)

    if node_name is None:
        try:
            return booth_env.config.read_raw()
        except RawFileError as e:
            report_processor.report(raw_file_error_report(e))
        if report_processor.has_errors:
            raise LibraryError()

    com_cmd = BoothGetConfig(env.report_processor, instance_name)
    com_cmd.set_targets(
        [env.get_node_target_factory().get_target_from_hostname(node_name)]
    )
    # pylint: disable=unsubscriptable-object
    # In general, pylint is right. And it cannot know in this case code is OK.
    # It is covered by tests.
    remote_data = run_and_raise(env.get_node_communicator(), com_cmd)[0][1]
    try:
        # TODO switch to new file transfer commands (not implemented yet)
        # which send and receive configs as bytes instead of strings
        return remote_data["config"]["data"].encode("utf-8")
    except KeyError as e:
        raise LibraryError(
            ReportItem.error(reports.messages.InvalidResponseFormat(node_name))
        ) from e


def config_ticket_add(
    env: LibraryEnvironment,
    ticket_name,
    options,
    instance_name=None,
    allow_unknown_options=False,
):
    """
    add a ticket to booth configuration

    env
    string ticket_name -- the name of the ticket to be created
    dict options -- options for the ticket
    string instance_name -- booth instance name
    bool allow_unknown_options -- allow using options unknown to pcs
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    try:
        booth_conf = booth_env.config.read_to_facade()
        options_pairs = validate.values_to_pairs(
            options, config_validators.ticket_options_normalization()
        )
        report_processor.report_list(
            config_validators.add_ticket(
                booth_conf,
                ticket_name,
                options_pairs,
                allow_unknown_options=allow_unknown_options,
            )
        )
        if report_processor.has_errors:
            raise LibraryError()
        booth_conf.add_ticket(
            ticket_name, validate.pairs_to_values(options_pairs)
        )
        booth_env.config.write_facade(booth_conf, can_overwrite=True)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()


def config_ticket_remove(
    env: LibraryEnvironment,
    ticket_name,
    instance_name=None,
):
    """
    remove a ticket from booth configuration

    env
    string ticket_name -- the name of the ticket to be removed
    string instance_name -- booth instance name
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    try:
        booth_conf = booth_env.config.read_to_facade()
        report_processor.report_list(
            config_validators.remove_ticket(booth_conf, ticket_name)
        )
        if report_processor.has_errors:
            raise LibraryError()
        booth_conf.remove_ticket(ticket_name)
        booth_env.config.write_facade(booth_conf, can_overwrite=True)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()


def create_in_cluster(
    env: LibraryEnvironment,
    ip: str,
    instance_name: Optional[str] = None,
    allow_absent_resource_agent: bool = False,
):
    """
    Create group with ip resource and booth resource

    env -- provides all for communication with externals
    ip -- float ip address for the operation of the booth
    instance_name -- booth instance name
    allow_absent_resource_agent -- allowing creating booth resource even
        if its agent is not installed
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    # Booth config path goes to CIB. Working with a mocked booth configs would
    # not work coorectly as the path would point to a mock file (the path to a
    # mock file is unknown to us in the lib anyway)
    # It makes sense to work with a mocked CIB, though. Users can do other
    # changes to the CIB and push them to the cluster at once.
    _ensure_live_booth_env(booth_env)
    resources_section = get_resources(env.get_cib())
    id_provider = IdProvider(resources_section)
    instance_name = booth_env.instance_name

    # validate
    if resource.find_for_config(resources_section, booth_env.config_path):
        report_processor.report(
            ReportItem.error(reports.messages.BoothAlreadyInCib(instance_name))
        )
    # verify the config exists and is readable
    try:
        booth_env.config.raw_file.read()
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    if report_processor.has_errors:
        raise LibraryError()
    # validation done

    create_id = partial(
        resource.create_resource_id, resources_section, instance_name
    )
    create_primitive = partial(
        primitive.create,
        env.report_processor,
        env.cmd_runner(),
        resources_section,
        id_provider,
        enable_agent_self_validation=False,
    )
    agent_factory = ResourceAgentFacadeFactory(
        env.cmd_runner(), report_processor
    )

    # Group id validation is not needed since create_id creates a new unique
    # booth group identifier
    hierarchy.move_resources_to_group(
        group.append_new(resources_section, create_id("group")),
        [
            create_primitive(
                create_id("ip"),
                _get_agent_facade(
                    env.report_processor,
                    agent_factory,
                    allow_absent_resource_agent,
                    ResourceAgentName("ocf", "heartbeat", "IPaddr2"),
                ),
                instance_attributes={"ip": ip},
            ),
            create_primitive(
                create_id("service"),
                _get_agent_facade(
                    env.report_processor,
                    agent_factory,
                    allow_absent_resource_agent,
                    ResourceAgentName("ocf", "pacemaker", "booth-site"),
                ),
                instance_attributes={"config": booth_env.config_path},
            ),
        ],
    )

    env.push_cib()


def remove_from_cluster(
    env: LibraryEnvironment,
    resource_remove,
    instance_name=None,
    allow_remove_multiple=False,
):
    """
    Remove group with ip resource and booth resource

    env -- provides all for communication with externals
    function resource_remove -- provisional hack til resources are moved to lib
    string instance_name -- booth instance name
    bool allow_remove_multiple -- remove all resources if more than one found
    """
    # TODO resource_remove is provisional hack til resources are moved to lib
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    # This command does not work with booth config files at all, let's reject
    # them then.
    _ensure_live_booth_env(booth_env)

    resource.get_remover(resource_remove)(
        _find_resource_elements_for_operation(
            report_processor,
            get_resources(env.get_cib()),
            booth_env,
            allow_remove_multiple,
        )
    )


def restart(
    env: LibraryEnvironment,
    resource_restart,
    instance_name=None,
    allow_multiple=False,
):
    """
    Restart group with ip resource and booth resource

    env -- provides all for communication with externals
    function resource_restart -- provisional hack til resources are moved to lib
    string instance_name -- booth instance name
    bool allow_remove_multiple -- remove all resources if more than one found
    """
    # TODO resource_remove is provisional hack til resources are moved to lib
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)

    for booth_element in _find_resource_elements_for_operation(
        report_processor,
        get_resources(env.get_cib()),
        booth_env,
        allow_multiple,
    ):
        resource_restart([booth_element.attrib["id"]])


def ticket_grant(
    env: LibraryEnvironment,
    ticket_name,
    site_ip=None,
    instance_name=None,
):
    """
    Grant a ticket to the site specified by site_ip

    env
    string ticket_name -- the name of the ticket to be granted
    string site_ip -- IP of the site to grant the ticket to, None for local
    string instance_name -- booth instance name
    """
    return _ticket_operation(
        "grant",
        env,
        ticket_name,
        site_ip=site_ip,
        instance_name=instance_name,
    )


def ticket_revoke(
    env: LibraryEnvironment,
    ticket_name,
    site_ip=None,
    instance_name=None,
):
    """
    Revoke a ticket from the site specified by site_ip

    env
    string ticket_name -- the name of the ticket to be revoked
    string site_ip -- IP of the site to revoke the ticket from, None for local
    string instance_name -- booth instance name
    """
    return _ticket_operation(
        "revoke",
        env,
        ticket_name,
        site_ip=site_ip,
        instance_name=instance_name,
    )


def _ticket_operation(
    operation, env: LibraryEnvironment, ticket_name, site_ip, instance_name
):
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)

    if not site_ip:
        site_ip_list = resource.find_bound_ip(
            get_resources(env.get_cib()), booth_env.config_path
        )
        if len(site_ip_list) != 1:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.BoothCannotDetermineLocalSiteIp()
                )
            )
        site_ip = site_ip_list[0]

    stdout, stderr, return_code = env.cmd_runner().run(
        [settings.booth_binary, operation, "-s", site_ip, ticket_name]
    )

    if return_code != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.BoothTicketOperationFailed(
                    operation,
                    join_multilines([stderr, stdout]),
                    site_ip,
                    ticket_name,
                )
            )
        )


def config_sync(
    env: LibraryEnvironment,
    instance_name=None,
    skip_offline_nodes=False,
):
    """
    Send specified local booth configuration to all nodes in the local cluster.

    env
    string instance_name -- booth instance name
    skip_offline_nodes -- if True offline nodes will be skipped
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    if not env.is_cib_live:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired([file_type_codes.CIB])
            )
        )

    cluster_nodes_names, report_list = get_existing_nodes_names(
        env.get_corosync_conf()
    )
    if not cluster_nodes_names:
        report_list.append(
            ReportItem.error(reports.messages.CorosyncConfigNoNodesDefined())
        )
    report_processor.report_list(report_list)

    try:
        booth_conf_data = booth_env.config.read_raw()
        booth_conf = booth_env.config.raw_to_facade(booth_conf_data)
        if isinstance(booth_env.config.raw_file, GhostFile):
            authfile_data = booth_env.key.read_raw()
            authfile_path = booth_conf.get_authfile()
            authfile_name = (
                os.path.basename(authfile_path) if authfile_path else None
            )
        else:
            (
                authfile_name,
                authfile_data,
                authfile_report_list,
            ) = config_files.get_authfile_name_and_data(booth_conf)
            report_processor.report_list(authfile_report_list)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()

    com_cmd = BoothSendConfig(
        env.report_processor,
        booth_env.instance_name,
        booth_conf_data,
        authfile=authfile_name,
        authfile_data=authfile_data,
        skip_offline_targets=skip_offline_nodes,
    )
    com_cmd.set_targets(
        env.get_node_target_factory().get_target_list(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
    )
    run_and_raise(env.get_node_communicator(), com_cmd)


def enable_booth(env: LibraryEnvironment, instance_name=None):
    """
    Enable specified instance of booth service, systemd systems supported only.

    env
    string instance_name -- booth instance name
    """
    ensure_is_systemd(env.service_manager)
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        env.service_manager.enable("booth", instance=instance_name)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    env.report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_ENABLE,
                "booth",
                instance=instance_name,
            )
        )
    )


def disable_booth(env: LibraryEnvironment, instance_name=None):
    """
    Disable specified instance of booth service, systemd systems supported only.

    env
    string instance_name -- booth instance name
    """
    ensure_is_systemd(env.service_manager)
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        env.service_manager.disable("booth", instance=instance_name)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    env.report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_DISABLE,
                "booth",
                instance=instance_name,
            )
        )
    )


def start_booth(env: LibraryEnvironment, instance_name=None):
    """
    Start specified instance of booth service, systemd systems supported only.
        On non-systemd systems it can be run like this:
        BOOTH_CONF_FILE=<booth-file-path> /etc/initd/booth-arbitrator

    env
    string instance_name -- booth instance name
    """
    ensure_is_systemd(env.service_manager)
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        env.service_manager.start("booth", instance=instance_name)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    env.report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_START,
                "booth",
                instance=instance_name,
            )
        )
    )


def stop_booth(env: LibraryEnvironment, instance_name=None):
    """
    Stop specified instance of booth service, systemd systems supported only.

    env
    string instance_name -- booth instance name
    """
    ensure_is_systemd(env.service_manager)
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        env.service_manager.stop("booth", instance=instance_name)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    env.report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_STOP,
                "booth",
                instance=instance_name,
            )
        )
    )


def pull_config(env: LibraryEnvironment, node_name, instance_name=None):
    """
    Get config from specified node and save it on local system. It will
    rewrite existing files.

    env
    string node_name -- name of the node from which the config should be fetched
    string instance_name -- booth instance name
    """
    report_processor = env.report_processor
    booth_env = env.get_booth_env(instance_name)
    instance_name = booth_env.instance_name
    _ensure_live_env(env, booth_env)
    conf_dir = os.path.dirname(booth_env.config_path)

    env.report_processor.report(
        ReportItem.info(
            reports.messages.BoothFetchingConfigFromNode(
                node_name,
                config=instance_name,
            )
        )
    )
    com_cmd = BoothGetConfig(env.report_processor, instance_name)
    com_cmd.set_targets(
        [env.get_node_target_factory().get_target_from_hostname(node_name)]
    )
    # pylint: disable=unsubscriptable-object
    # In general, pylint is right. And it cannot know in this case code is OK.
    # It is covered by tests.
    output = run_and_raise(env.get_node_communicator(), com_cmd)[0][1]
    try:
        # TODO adapt to new file transfer framework once it is written
        if (
            output["authfile"]["name"] is not None
            and output["authfile"]["data"]
        ):
            authfile_name = output["authfile"]["name"]
            report_list = config_validators.check_instance_name(authfile_name)
            if report_list:
                raise LibraryError(*report_list)
            booth_key = FileInstance.for_booth_key(authfile_name)
            booth_key.write_raw(
                base64.b64decode(output["authfile"]["data"].encode("utf-8")),
                can_overwrite=True,
            )
        booth_env.config.write_raw(
            output["config"]["data"].encode("utf-8"), can_overwrite=True
        )
        env.report_processor.report(
            ReportItem.info(
                reports.messages.BoothConfigAcceptedByNode(
                    name_list=[instance_name]
                )
            )
        )
    except RawFileError as e:
        if not os.path.exists(conf_dir):
            report_processor.report(
                ReportItem.error(reports.messages.BoothPathNotExists(conf_dir))
            )
        else:
            report_processor.report(raw_file_error_report(e))
    except KeyError as e:
        raise LibraryError(
            ReportItem.error(reports.messages.InvalidResponseFormat(node_name))
        ) from e
    if report_processor.has_errors:
        raise LibraryError()


def get_status(env: LibraryEnvironment, instance_name=None):
    """
    get booth status info

    env
    string instance_name -- booth instance name
    """
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name
    report_msg = status.check_authfile_misconfiguration(
        booth_env, env.report_processor
    )
    if report_msg:
        env.report_processor.report(reports.ReportItem.warning(report_msg))
    return {
        "status": status.get_daemon_status(env.cmd_runner(), instance_name),
        "ticket": status.get_tickets_status(env.cmd_runner(), instance_name),
        "peers": status.get_peers_status(env.cmd_runner(), instance_name),
    }


def _find_resource_elements_for_operation(
    report_processor: ReportProcessor,
    resources_section,
    booth_env,
    allow_multiple,
):
    booth_element_list = resource.find_for_config(
        resources_section,
        booth_env.config_path,
    )

    if not booth_element_list:
        report_processor.report(
            ReportItem.error(
                reports.messages.BoothNotExistsInCib(booth_env.instance_name)
            )
        )
    elif len(booth_element_list) > 1:
        report_processor.report(
            ReportItem(
                severity=get_severity(
                    report_codes.FORCE,
                    allow_multiple,
                ),
                message=reports.messages.BoothMultipleTimesInCib(
                    booth_env.instance_name,
                ),
            )
        )
    if report_processor.has_errors:
        raise LibraryError()

    return booth_element_list


def _ensure_live_booth_env(booth_env):
    if booth_env.ghost_file_codes:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired(
                    booth_env.ghost_file_codes
                )
            )
        )


def _ensure_live_env(env: LibraryEnvironment, booth_env):
    not_live = (
        booth_env.ghost_file_codes
        +
        # parenthesis are cruciual, otherwise the if..else influences
        # booth_env.ghost_file_codes as well
        ([file_type_codes.CIB] if not env.is_cib_live else [])
    )
    if not_live:
        raise LibraryError(
            ReportItem.error(reports.messages.LiveEnvironmentRequired(not_live))
        )


def _get_agent_facade(
    report_processor: reports.ReportProcessor,
    factory: ResourceAgentFacadeFactory,
    allow_absent_agent: bool,
    name: ResourceAgentName,
) -> ResourceAgentFacade:
    try:
        return factory.facade_from_parsed_name(name)
    except UnableToGetAgentMetadata as e:
        if allow_absent_agent:
            report_processor.report(
                resource_agent_error_to_report_item(
                    e, reports.ReportItemSeverity.warning()
                )
            )
            return factory.void_facade_from_parsed_name(name)
        report_processor.report(
            resource_agent_error_to_report_item(
                e, reports.ReportItemSeverity.error(reports.codes.FORCE)
            )
        )
        raise LibraryError() from e
    except ResourceAgentError as e:
        report_processor.report(
            resource_agent_error_to_report_item(
                e, reports.ReportItemSeverity.error()
            )
        )
        raise LibraryError() from e



from collections import namedtuple
from typing import (
    List,
    Optional,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.lib.validate import validate_add_remove_items

from . import group
from .bundle import is_bundle
from .clone import (
    get_parent_any_clone,
    is_any_clone,
    is_master,
    is_promotable_clone,
)
from .common import (
    get_inner_resources,
    get_parent_resource,
    is_resource,
    is_wrapper_resource,
)
from .stonith import is_stonith


def validate_move_resources_to_group(
    group_element: _Element,
    resource_element_list: List[_Element],
    adjacent_resource_element: Optional[_Element],
) -> reports.ReportItemList:
    """
    Validates that existing resources can be moved into a group,
    optionally beside an adjacent_resource_element

    group_element -- the group to put resources into
    resource_element_list -- resources that are being moved into the group
    adjacent_resource_element -- put resources beside this one
    """
    report_list: reports.ReportItemList = []

    # Validate types of resources and their parents
    for resource_element in resource_element_list:
        # Only primitive resources can be moved
        if not is_resource(resource_element):
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.IdBelongsToUnexpectedType(
                        str(resource_element.attrib["id"]),
                        ["primitive"],
                        resource_element.tag,
                    )
                )
            )
        elif is_wrapper_resource(resource_element):
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.CannotGroupResourceWrongType(
                        str(resource_element.attrib["id"]),
                        resource_element.tag,
                        parent_id=None,
                        parent_type=None,
                    )
                )
            )
        else:
            parent = get_parent_resource(resource_element)
            if parent is not None and not group.is_group(parent):
                # At the moment, moving resources out of bundles and clones
                # (or masters) is not possible
                report_list.append(
                    reports.ReportItem.error(
                        reports.messages.CannotGroupResourceWrongType(
                            str(resource_element.attrib["id"]),
                            resource_element.tag,
                            parent_id=str(parent.attrib["id"]),
                            parent_type=parent.tag,
                        )
                    )
                )

    # Validate that elements can be added
    # Check if the group element is a group
    if group.is_group(group_element):
        report_list += validate_add_remove_items(
            [str(resource.attrib["id"]) for resource in resource_element_list],
            [],
            [
                str(resource.attrib["id"])
                for resource in get_inner_resources(group_element)
            ],
            reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
            reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
            str(group_element.attrib["id"]),
            str(adjacent_resource_element.attrib["id"])
            if adjacent_resource_element is not None
            else None,
            True,
        )
    else:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.IdBelongsToUnexpectedType(
                    str(group_element.attrib["id"]),
                    expected_types=[group.TAG],
                    current_type=group_element.tag,
                )
            )
        )

    # Elements can always be removed from their old groups, except when the last
    # resource is removed but that is handled in resource.group_add for now, no
    # need to run the validation for removing elements
    return report_list


def validate_move(resource_element, master):
    """
    Validate moving a resource to a node

    etree resource_element -- the resource to be moved
    bool master -- limit moving to the master role
    """
    report_list = []

    if is_stonith(resource_element):
        report_list.append(
            reports.ReportItem.deprecation(
                reports.messages.ResourceStonithCommandsMismatch(
                    "stonith device"
                )
            )
        )

    analysis = _validate_move_ban_clear_analyzer(resource_element)

    if analysis.is_bundle:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.CannotMoveResourceBundle(
                    resource_element.get("id")
                )
            )
        )
        return report_list

    if (analysis.is_clone or analysis.is_in_clone) and not (
        analysis.is_promotable_clone or analysis.is_in_promotable_clone
    ):
        report_list.append(
            reports.ReportItem.error(
                reports.messages.CannotMoveResourceClone(
                    resource_element.get("id")
                )
            )
        )
        return report_list

    # Allow moving both promoted and demoted instances of promotable clones
    # (analysis.is_promotable_clone). Do not allow to move promoted instances
    # of promotables' inner resources (analysis.is_in_promotable_clone) as that
    # would create a constraint for the promoted role of a group or a primitive
    # which would not make sense if the group or primitive is later moved out
    # of its promotable clone. To be consistent, we do not allow to move
    # demoted instances of promotables' inner resources either. See
    # rhbz#1875301 for details.
    if not master and analysis.is_in_promotable_clone:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.CannotMoveResourcePromotableInner(
                    resource_element.get("id"),
                    analysis.promotable_clone_id,
                )
            )
        )
    elif master and not analysis.is_promotable_clone:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.CannotMoveResourceMasterResourceNotPromotable(
                    resource_element.get("id"),
                    promotable_id=analysis.promotable_clone_id,
                )
            )
        )

    return report_list


def validate_ban(resource_element, master):
    """
    Validate banning a resource on a node

    etree resource_element -- the resource to be banned
    bool master -- limit banning to the master role
    """
    report_list = []

    if is_stonith(resource_element):
        report_list.append(
            reports.ReportItem.deprecation(
                reports.messages.ResourceStonithCommandsMismatch(
                    "stonith device"
                )
            )
        )

    analysis = _validate_move_ban_clear_analyzer(resource_element)

    if master and not analysis.is_promotable_clone:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.CannotBanResourceMasterResourceNotPromotable(
                    resource_element.get("id"),
                    promotable_id=analysis.promotable_clone_id,
                )
            )
        )

    return report_list


def validate_unmove_unban(resource_element, master):
    """
    Validate unmoving/unbanning a resource to/on nodes

    etree resource_element -- the resource to be unmoved/unbanned
    bool master -- limit unmoving/unbanning to the master role
    """
    report_list = []

    if is_stonith(resource_element):
        report_list.append(
            reports.ReportItem.deprecation(
                reports.messages.ResourceStonithCommandsMismatch(
                    "stonith device"
                )
            )
        )

    analysis = _validate_move_ban_clear_analyzer(resource_element)

    if master and not analysis.is_promotable_clone:
        # pylint: disable=line-too-long
        report_list.append(
            reports.ReportItem.error(
                reports.messages.CannotUnmoveUnbanResourceMasterResourceNotPromotable(
                    resource_element.get("id"),
                    promotable_id=analysis.promotable_clone_id,
                )
            )
        )

    return report_list


class _MoveBanClearAnalysis(
    namedtuple(
        "_MoveBanClearAnalysis",
        [
            "is_bundle",
            "is_clone",
            "is_in_clone",
            "is_promotable_clone",
            "is_in_promotable_clone",
            "promotable_clone_id",
        ],
    )
):
    pass


def _validate_move_ban_clear_analyzer(resource_element):
    resource_is_bundle = False
    resource_is_clone = False
    resource_is_in_clone = False
    resource_is_promotable_clone = False
    resource_is_in_promotable_clone = False
    promotable_clone_element = None

    if is_bundle(resource_element):
        resource_is_bundle = True
    elif is_any_clone(resource_element):
        resource_is_clone = True
        if is_master(resource_element) or is_promotable_clone(resource_element):
            resource_is_promotable_clone = True
            promotable_clone_element = resource_element
    elif get_parent_any_clone(resource_element) is not None:
        parent_clone = get_parent_any_clone(resource_element)
        resource_is_in_clone = True
        if is_master(parent_clone) or is_promotable_clone(parent_clone):
            resource_is_in_promotable_clone = True
            promotable_clone_element = parent_clone
    return _MoveBanClearAnalysis(
        resource_is_bundle,
        resource_is_clone,
        resource_is_in_clone,
        resource_is_promotable_clone,
        resource_is_in_promotable_clone,
        (
            promotable_clone_element.get("id")
            if promotable_clone_element is not None
            else None
        ),
    )



from pcs import (
    qdevice,
    usage,
)
from pcs.cli.common.routing import create_router

qdevice_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.qdevice(argv)),
        "status": qdevice.qdevice_status_cmd,
        "setup": qdevice.qdevice_setup_cmd,
        "destroy": qdevice.qdevice_destroy_cmd,
        "start": qdevice.qdevice_start_cmd,
        "stop": qdevice.qdevice_stop_cmd,
        "kill": qdevice.qdevice_kill_cmd,
        "enable": qdevice.qdevice_enable_cmd,
        "disable": qdevice.qdevice_disable_cmd,
        # following commands are internal use only, called from pcsd
        "sign-net-cert-request": qdevice.qdevice_sign_net_cert_request_cmd,
        "net-client": create_router(
            {
                "setup": qdevice.qdevice_net_client_setup_cmd,
                "import-certificate": (
                    qdevice.qdevice_net_client_import_certificate_cmd
                ),
                "destroy": qdevice.qdevice_net_client_destroy,
            },
            ["qdevice", "net-client"],
        ),
    },
    ["qdevice"],
)



from unittest import TestCase


class SendAllConfigToNodeTest(TestCase):
    """
    tested in:
        pcs_test.tier0.lib.commands.cluster.test_add_nodes.AddNodeFull
        pcs_test.tier0.lib.commands.cluster.test_add_nodes
            .FailureBoothConfigsDistribution
    """



from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.tag import command
from pcs.lib.errors import LibraryError

from pcs_test.tools.misc import dict_to_modifiers


class TagCreate(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["tag"])
        self.tag = mock.Mock(spec_set=["create"])
        self.lib.tag = self.tag

    def _call_cmd(self, argv):
        command.tag_create(self.lib, argv, dict_to_modifiers({}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.tag.create.assert_not_called()

    def test_not_enough_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg1"])
        self.assertIsNone(cm.exception.message)
        self.tag.create.assert_not_called()

    def test_minimum_args(self):
        self._call_cmd(["arg1", "arg2"])
        self.tag.create.assert_called_once_with("arg1", ["arg2"])

    def test_empty_args(self):
        self._call_cmd(["", ""])
        self.tag.create.assert_called_once_with("", [""])

    def test_multiple_args(self):
        self._call_cmd(["tagid", "id1", "id2"])
        self.tag.create.assert_called_once_with("tagid", ["id1", "id2"])


@mock.patch("pcs.cli.tag.command.print")
class TagConfig(TestCase):
    tag_dicts = [
        {
            "tag_id": "tag1",
            "idref_list": ["i1", "i2", "i3"],
        },
        {
            "tag_id": "tag2",
            "idref_list": ["j1", "j2", "j3"],
        },
    ]

    def setUp(self):
        self.lib = mock.Mock(spec_set=["tag"])
        self.tag = mock.Mock(spec_set=["config"])
        self.lib.tag = self.tag

    def _call_cmd(self, argv):
        command.tag_config(self.lib, argv, dict_to_modifiers({}))

    @mock.patch("pcs.cli.tag.command.print_to_stderr")
    def test_no_args_no_tags(self, mock_stderr_print, mock_print):
        self.tag.config.return_value = []
        self._call_cmd([])
        self.tag.config.assert_called_once_with([])
        mock_print.assert_not_called()
        mock_stderr_print.assert_called_once_with(" No tags defined")

    def test_no_args_all_tags(self, mock_print):
        self.tag.config.return_value = self.tag_dicts
        self._call_cmd([])
        self.tag.config.assert_called_once_with([])
        mock_print.assert_called_once_with(
            dedent(
                """\
            tag1
              i1
              i2
              i3
            tag2
              j1
              j2
              j3"""
            )
        )

    def test_specified_tag(self, mock_print):
        self.tag.config.return_value = self.tag_dicts[1:2]
        self._call_cmd(["tag2"])
        self.tag.config.assert_called_once_with(["tag2"])
        mock_print.assert_called_once_with(
            dedent(
                """\
            tag2
              j1
              j2
              j3"""
            )
        )

    def test_specified_another_tag(self, mock_print):
        self.tag.config.return_value = self.tag_dicts[0:1]
        self._call_cmd(["tag1"])
        self.tag.config.assert_called_once_with(["tag1"])
        mock_print.assert_called_once_with(
            dedent(
                """\
            tag1
              i1
              i2
              i3"""
            )
        )

    def test_specified_more_tags_are_printed_in_given_order(self, mock_print):
        self.tag.config.return_value = self.tag_dicts[::-1]
        self._call_cmd(["tag2", "tag1"])
        self.tag.config.assert_called_once_with(["tag2", "tag1"])
        mock_print.assert_called_once_with(
            dedent(
                """\
            tag2
              j1
              j2
              j3
            tag1
              i1
              i2
              i3"""
            )
        )

    def test_no_print_on_exception(self, mock_print):
        self.tag.config.side_effect = LibraryError()
        with self.assertRaises(LibraryError):
            self._call_cmd(["something"])
        self.tag.config.assert_called_once_with(["something"])
        mock_print.assert_not_called()


class TagRemove(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["tag"])
        self.tag = mock.Mock(spec_set=["remove"])
        self.lib.tag = self.tag

    def _call_cmd(self, argv):
        command.tag_remove(self.lib, argv, dict_to_modifiers({}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.tag.remove.assert_not_called()

    def test_minimum_args(self):
        self._call_cmd(["arg1"])
        self.tag.remove.assert_called_once_with(["arg1"])

    def test_more_args(self):
        self._call_cmd(["arg1", "arg2"])
        self.tag.remove.assert_called_once_with(["arg1", "arg2"])


class TagUpdate(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["tag"])
        self.tag = mock.Mock(spec_set=["update"])
        self.lib.tag = self.tag

    def _call_cmd(self, argv, options=None):
        if options is None:
            options = {}
        command.tag_update(self.lib, argv, dict_to_modifiers(options))

    def assert_exception_hint(self, argv):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(argv)
        self.assertEqual(
            cm.exception.hint,
            "Specify at least one id for 'add' or 'remove' arguments.",
        )
        self.tag.update.assert_not_called()

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.tag.update.assert_not_called()

    def test_no_add_remove_keywords(self):
        self.assert_exception_hint(["tag_id"])

    def test_add_without_ids(self):
        self.assert_exception_hint(["tag_id", "add"])

    def test_remove_without_ids(self):
        self.assert_exception_hint(["tag_id", "remove"])

    def test_add_remove_without_ids(self):
        self.assert_exception_hint(["tag_id", "add", "remove"])

    def test_add_and_empty_remove(self):
        self.assert_exception_hint(["tag_id", "add", "id1", "remove"])

    def test_remove_and_empty_add(self):
        self.assert_exception_hint(["tag_id", "add", "remove", "id1"])

    def test_both_after_and_before(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(
                ["tag_id", "add", "id1"],
                dict(after="A", before="B"),
            )
        self.assertEqual(
            cm.exception.message, "Cannot specify both --before and --after"
        )
        self.tag.update.assert_not_called()

    def test_only_add(self):
        self._call_cmd(["tag_id", "add", "id1"])
        self.tag.update.assert_called_once_with(
            "tag_id",
            ["id1"],
            [],
            adjacent_idref=None,
            put_after_adjacent=True,
        )

    def test_only_remove(self):
        self._call_cmd(["tag_id", "remove", "id1"])
        self.tag.update.assert_called_once_with(
            "tag_id",
            [],
            ["id1"],
            adjacent_idref=None,
            put_after_adjacent=True,
        )

    def test_both_add_remove(self):
        self._call_cmd(["tag_id", "remove", "i", "j", "add", "k", "l"])
        self.tag.update.assert_called_once_with(
            "tag_id",
            ["k", "l"],
            ["i", "j"],
            adjacent_idref=None,
            put_after_adjacent=True,
        )

    def test_only_add_after(self):
        self._call_cmd(["tag_id", "add", "id1"], dict(after="A"))
        self.tag.update.assert_called_once_with(
            "tag_id",
            ["id1"],
            [],
            adjacent_idref="A",
            put_after_adjacent=True,
        )

    def test_only_add_before(self):
        self._call_cmd(["tag_id", "add", "id1"], dict(before="B"))
        self.tag.update.assert_called_once_with(
            "tag_id",
            ["id1"],
            [],
            adjacent_idref="B",
            put_after_adjacent=False,
        )

    def test_add_after_and_remove(self):
        self._call_cmd(
            ["tag_id", "add", "id1", "id2", "remove", "id3", "id4"],
            dict(after="A"),
        )
        self.tag.update.assert_called_once_with(
            "tag_id",
            ["id1", "id2"],
            ["id3", "id4"],
            adjacent_idref="A",
            put_after_adjacent=True,
        )

    def test_add_before_and_remove(self):
        self._call_cmd(
            ["tag_id", "add", "id1", "id2", "remove", "id3", "id4"],
            dict(before="B"),
        )
        self.tag.update.assert_called_once_with(
            "tag_id",
            ["id1", "id2"],
            ["id3", "id4"],
            adjacent_idref="B",
            put_after_adjacent=False,
        )



import threading
import uuid
from collections import namedtuple
from typing import (
    MutableSet,
    Optional,
    TypeVar,
    Union,
)

from lxml import etree
from lxml.etree import _Element

from pcs.common.types import StringCollection

T = TypeVar("T", bound=type)


def get_all_subclasses(cls: T) -> MutableSet[T]:
    subclasses = set(cls.__subclasses__())
    return subclasses.union(
        {s for c in subclasses for s in get_all_subclasses(c)}
    )


def get_unique_uuid(already_used: StringCollection) -> str:
    is_duplicate = True
    while is_duplicate:
        candidate = str(uuid.uuid4())
        is_duplicate = candidate in already_used
    return candidate


def run_parallel(worker, data_list):
    thread_list = []
    for args, kwargs in data_list:
        thread = threading.Thread(target=worker, args=args, kwargs=kwargs)
        thread.daemon = True
        thread_list.append(thread)
        thread.start()

    for thread in thread_list:
        thread.join()


def format_environment_error(e):
    return format_os_error(e)


def format_os_error(e: OSError):
    if e.filename:
        return "{0}: '{1}'".format(e.strerror, e.filename)
    return e.strerror


def xml_fromstring(xml: str) -> _Element:
    # If the xml contains encoding declaration such as:
    # <?xml version="1.0" encoding="UTF-8"?>
    # we get an exception in python3:
    # ValueError: Unicode strings with encoding declaration are not supported.
    # Please use bytes input or XML fragments without declaration.
    # So we encode the string to bytes.
    return etree.fromstring(
        xml.encode("utf-8"),
        # it raises on a huge xml without the flag huge_tree=True
        # see https://bugzilla.redhat.com/show_bug.cgi?id=1506864
        etree.XMLParser(huge_tree=True),
    )


def timeout_to_seconds(timeout: Union[int, str]) -> Optional[int]:
    """
    Transform pacemaker style timeout to number of seconds. If `timeout` is not
    a valid timeout, `None` is returned.

    timeout -- timeout string
    """
    try:
        candidate = int(timeout)
        if candidate >= 0:
            return candidate
        return None
    except ValueError:
        pass
    # Now we know the timeout is not an integer nor an integer string.
    # Let's make sure mypy knows the timeout is a string as well.
    timeout = str(timeout)
    suffix_multiplier = {
        "s": 1,
        "sec": 1,
        "m": 60,
        "min": 60,
        "h": 3600,
        "hr": 3600,
    }
    for suffix, multiplier in suffix_multiplier.items():
        if timeout.endswith(suffix) and timeout[: -len(suffix)].isdigit():
            return int(timeout[: -len(suffix)]) * multiplier
    return None


class Version(namedtuple("Version", ["major", "minor", "revision"])):
    def __new__(
        cls,
        major: int,
        minor: Optional[int] = None,
        revision: Optional[int] = None,
    ):
        return super(Version, cls).__new__(cls, major, minor, revision)

    @property
    def as_full_tuple(self):
        return (
            self.major,
            self.minor if self.minor is not None else 0,
            self.revision if self.revision is not None else 0,
        )

    def normalize(self):
        return self.__class__(*self.as_full_tuple)

    def __str__(self):
        return ".".join([str(x) for x in self if x is not None])

    def __lt__(self, other):
        return self.as_full_tuple < other.as_full_tuple

    def __le__(self, other):
        return self.as_full_tuple <= other.as_full_tuple

    def __eq__(self, other):
        return self.as_full_tuple == other.as_full_tuple

    def __ne__(self, other):
        return self.as_full_tuple != other.as_full_tuple

    def __gt__(self, other):
        return self.as_full_tuple > other.as_full_tuple

    def __ge__(self, other):
        return self.as_full_tuple >= other.as_full_tuple



import logging
import multiprocessing as mp
import os

WORKER_LOGGER = "pcs_worker"


class Logger(logging.Logger):
    # pylint: disable=too-many-arguments
    def makeRecord(  # type: ignore
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func=None,
        extra=None,
        sinfo=None,
    ) -> logging.LogRecord:
        pid = os.getpid()
        return super().makeRecord(
            name,
            level,
            fn,
            lno,
            f"Worker#{pid}: {msg}",
            args,
            exc_info,
            func=func,
            extra=extra,
            sinfo=sinfo,
        )


def setup_worker_logger(queue: mp.Queue) -> logging.Logger:
    """
    Creates and configures worker's logger
    :return: Logger instance
    """
    logging.setLoggerClass(Logger)
    logger = logging.getLogger(WORKER_LOGGER)
    logger.setLevel(logging.DEBUG)

    queue_handler = logging.handlers.QueueHandler(queue)  # type: ignore
    logger.addHandler(queue_handler)

    return logger



from enum import auto

from pcs.common.types import AutoNameEnum


class TaskFinishType(AutoNameEnum):
    UNFINISHED = auto()
    UNHANDLED_EXCEPTION = auto()
    FAIL = auto()
    SUCCESS = auto()
    KILL = auto()


class TaskState(AutoNameEnum):
    CREATED = auto()
    QUEUED = auto()
    EXECUTED = auto()
    FINISHED = auto()


class TaskKillReason(AutoNameEnum):
    USER = auto()
    COMPLETION_TIMEOUT = auto()
    INTERNAL_MESSAGING_ERROR = auto()



from typing import (
    Iterable,
    Mapping,
    Optional,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.types import StringCollection
from pcs.lib.cib.node import PacemakerNode
from pcs.lib.cib.resource import primitive
from pcs.lib.cib.resource.types import ResourceOperationIn
from pcs.lib.cib.tools import IdProvider
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent import (
    ResourceAgentFacade,
    ResourceAgentMetadata,
    ResourceAgentName,
)

AGENT_NAME = ResourceAgentName("ocf", "pacemaker", "remote")


_IS_REMOTE_AGENT_XPATH_SNIPPET = """
    @class="{0}" and @provider="{1}" and @type="{2}"
""".format(
    AGENT_NAME.standard, AGENT_NAME.provider, AGENT_NAME.type
)

_HAS_SERVER_XPATH_SNIPPET = """
    instance_attributes/nvpair[
        @name="server"
        and
        string-length(@value) > 0
    ]
"""


def find_node_list(tree):
    """
    Return list of remote nodes from the specified element tree

    etree.Element tree -- an element to search remote nodes in
    """
    node_list = [
        PacemakerNode(
            nvpair.getparent().getparent().attrib["id"], nvpair.attrib["value"]
        )
        for nvpair in tree.xpath(
            ".//primitive[{is_remote}]/{has_server}".format(
                is_remote=_IS_REMOTE_AGENT_XPATH_SNIPPET,
                has_server=_HAS_SERVER_XPATH_SNIPPET,
            )
        )
    ]

    node_list.extend(
        [
            PacemakerNode(primitive.attrib["id"], primitive.attrib["id"])
            for primitive in tree.xpath(
                ".//primitive[{is_remote} and not({has_server})]".format(
                    is_remote=_IS_REMOTE_AGENT_XPATH_SNIPPET,
                    has_server=_HAS_SERVER_XPATH_SNIPPET,
                )
            )
        ]
    )

    return node_list


def find_node_resources(resources_section, node_identifier):
    """
    Return list of resource elements that match to node_identifier

    etree.Element resources_section is a search element
    string node_identifier could be id of the resource or its instance attribute
        "server"
    """
    return resources_section.xpath(
        f"""
            .//primitive[
                {_IS_REMOTE_AGENT_XPATH_SNIPPET} and (
                    @id=$identifier
                    or
                    instance_attributes/nvpair[
                        @name="server"
                        and
                        @value=$identifier
                    ]
                )
            ]
        """,
        identifier=node_identifier,
    )


def get_node_name_from_resource(resource_element):
    """
    Return the node name from a remote node resource, None for other resources

    etree.Element resource_element
    """
    if not (
        resource_element.attrib.get("class", "") == AGENT_NAME.standard
        and resource_element.attrib.get("provider", "") == AGENT_NAME.provider
        and resource_element.attrib.get("type", "") == AGENT_NAME.type
    ):
        return None
    return resource_element.attrib["id"]


def _validate_server_not_used(
    agent: ResourceAgentMetadata, option_dict: Mapping[str, str]
) -> reports.ReportItemList:
    if "server" in option_dict:
        return [
            reports.ReportItem.error(
                reports.messages.InvalidOptions(
                    ["server"],
                    sorted(
                        [
                            attr.name
                            for attr in agent.parameters
                            if attr.name != "server"
                        ]
                    ),
                    "resource",
                ),
            )
        ]
    return []


def validate_host_not_conflicts(
    existing_nodes_addrs: StringCollection,
    node_name: str,
    instance_attributes: Mapping[str, str],
) -> reports.ReportItemList:
    host = instance_attributes.get("server", node_name)
    if host in existing_nodes_addrs:
        return [
            reports.ReportItem.error(reports.messages.IdAlreadyExists(host))
        ]
    return []


def validate_create(
    existing_nodes_names: StringCollection,
    existing_nodes_addrs: StringCollection,
    resource_agent: ResourceAgentMetadata,
    new_node_name: str,
    new_node_addr: str,
    instance_attributes: Mapping[str, str],
) -> reports.ReportItemList:
    """
    validate inputs for create

    existing_nodes_names -- node names already in use
    existing_nodes_addrs -- node addresses already in use
    ResourceAgent resource_agent -- pacemaker_remote resource agent
    new_node_name -- the name of the future node
    new_node_addr -- the address of the future node
    instance_attributes -- data for the future resource instance attributes
    """
    report_list = _validate_server_not_used(resource_agent, instance_attributes)

    addr_is_used = False
    if new_node_addr in existing_nodes_addrs:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.IdAlreadyExists(new_node_addr)
            )
        )
        addr_is_used = True

    if not addr_is_used or new_node_addr != new_node_name:
        if new_node_name in existing_nodes_names:
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.IdAlreadyExists(new_node_name)
                )
            )

    return report_list


def _prepare_instance_attributes(
    instance_attributes: Optional[Mapping[str, str]],
    host: str,
) -> Mapping[str, str]:
    enriched_instance_attributes = (
        dict(instance_attributes) if instance_attributes else {}
    )
    enriched_instance_attributes["server"] = host
    return enriched_instance_attributes


def create(
    report_processor: reports.ReportProcessor,
    cmd_runner: CommandRunner,
    resource_agent_facade: ResourceAgentFacade,
    resources_section: _Element,
    id_provider: IdProvider,
    host: str,
    node_name: str,
    raw_operation_list: Optional[Iterable[ResourceOperationIn]] = None,
    meta_attributes: Optional[Mapping[str, str]] = None,
    instance_attributes: Optional[Mapping[str, str]] = None,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
):
    # pylint: disable=too-many-arguments
    """
    Prepare all parts of remote resource and append it into the cib.

    report_processor -- tool for warning/info/error reporting
    resources_section -- place where new element will be appended
    node_name -- name of the remote node and id of new resource as well
    raw_operation_list -- specifies operations of resource
    meta_attributes -- specifies meta attributes of resource
    instance_attributes -- specifies instance attributes of resource
    allow_invalid_operation -- flag for skipping validation of operations
    allow_invalid_instance_attributes -- flag for skipping validation of
        instance_attributes
    use_default_operations -- flag for completion operations with default
        actions specified in resource agent
    """
    all_instance_attributes = _prepare_instance_attributes(
        instance_attributes, host
    )
    return primitive.create(
        report_processor,
        cmd_runner,
        resources_section,
        id_provider,
        node_name,
        resource_agent_facade,
        raw_operation_list,
        meta_attributes,
        all_instance_attributes,
        allow_invalid_operation,
        allow_invalid_instance_attributes,
        use_default_operations,
        # TODO remove this dirty fix
        # This is for passing info from the top (here) through another library
        # command (primitive.create) to instance attributes validator in
        # resource_agent. We handle the "server" attribute ourselves in this
        # command so we want to make sure it is not reported as an allowed
        # option.
        # How to fix:
        # 1) do not call one lib command from another
        # 2) split validation and cib modification in primitive.create
        # 3) call the validation from here and handle the results or config
        #    the validator before / when running it
        do_not_report_instance_attribute_server_exists=True,
        enable_agent_self_validation=False,
    )



from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.resource_agent.dto import ResourceAgentNameDto

from .operations import CibResourceOperationDto


@dataclass(frozen=True)
class CibResourcePrimitiveDto(DataTransferObject):
    id: str  # pylint: disable=invalid-name
    agent_name: ResourceAgentNameDto
    description: Optional[str]
    operations: Sequence[CibResourceOperationDto]
    meta_attributes: Sequence[CibNvsetDto]
    instance_attributes: Sequence[CibNvsetDto]
    utilization: Sequence[CibNvsetDto]



import re
from collections import Counter

from pcs.common import reports
from pcs.lib import validate
from pcs.lib.booth import constants

__TICKET_NAME_RE = re.compile(r"^[a-zA-Z0-9-]{1,63}$")


def check_instance_name(name):
    """
    Check that specified booth instance name is valid

    string name -- booth instance name
    """
    report_list = []
    if "/" in name:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.BoothInvalidName(
                    name, forbidden_characters="/"
                )
            )
        )
    return report_list


def create(site_list, arbitrator_list):
    """
    Validate creating a minimal booth config

    iterable site_list -- list of booth sites' addresses
    iterable arbitrator_list -- list of arbitrators' addresses
    """
    report_list = []
    peer_list = site_list + arbitrator_list

    if len(site_list) < 2:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.BoothLackOfSites(sorted(site_list))
            )
        )

    if len(peer_list) % 2 == 0:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.BoothEvenPeersNumber(len(peer_list))
            )
        )

    duplicate_addresses = {
        address for address, count in Counter(peer_list).items() if count > 1
    }
    if duplicate_addresses:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.BoothAddressDuplication(
                    sorted(duplicate_addresses)
                )
            )
        )

    return report_list


def add_ticket(
    conf_facade, ticket_name, ticket_options, allow_unknown_options=False
):
    """
    Validate adding a ticket to an existing booth config

    pcs.lib.booth.config_facade.ConfigFacade conf_facade -- a booth config
    string ticket_name -- the name of the ticket
    dict ticket_options -- ticket options
    bool allow_unknown_options -- if True, report unknown options as warnings
        instead of errors
    """
    return (
        validate_ticket_name(ticket_name)
        + _validate_ticket_unique(conf_facade, ticket_name)
        + _validate_ticket_options(ticket_options, allow_unknown_options)
    )


def remove_ticket(conf_facade, ticket_name):
    """
    Validate removing a ticket from an existing booth config

    pcs.lib.booth.config_facade.ConfigFacade conf_facade -- a booth config
    string ticket_name -- the name of the ticket
    """
    if not conf_facade.has_ticket(ticket_name):
        return [
            reports.ReportItem.error(
                reports.messages.BoothTicketDoesNotExist(ticket_name)
            )
        ]
    return []


def ticket_options_normalization() -> validate.TypeNormalizeFunc:
    return validate.option_value_normalization(
        {
            "mode": (
                lambda value: value.lower() if isinstance(value, str) else value
            )
        }
    )


def validate_ticket_name(ticket_name: str) -> reports.ReportItemList:
    if not __TICKET_NAME_RE.search(ticket_name):
        return [
            reports.ReportItem.error(
                reports.messages.BoothTicketNameInvalid(ticket_name)
            )
        ]
    return []


def _validate_ticket_unique(conf_facade, ticket_name):
    if conf_facade.has_ticket(ticket_name):
        return [
            reports.ReportItem.error(
                reports.messages.BoothTicketDuplicate(ticket_name)
            )
        ]
    return []


def _validate_ticket_options(options, allow_unknown_options):
    severity = reports.item.get_severity(
        reports.codes.FORCE, allow_unknown_options
    )
    validator_list = (
        [
            validate.NamesIn(
                constants.TICKET_KEYS,
                option_type="booth ticket",
                banned_name_list=constants.GLOBAL_KEYS,
                severity=severity,
            ),
        ]
        + [
            validate.ValueNotEmpty(option, None)
            for option in options
            if option != "mode"
        ]
        + [validate.ValueIn("mode", ["auto", "manual"], severity=severity)]
    )
    normalized_options = validate.values_to_pairs(
        options, lambda key, value: value.strip()
    )
    return validate.ValidatorAll(validator_list).validate(normalized_options)



from unittest import TestCase

from pcs.common import reports
from pcs.lib.commands import tag as cmd_tag

from pcs_test.tier0.lib.commands.tag.tag_common import (
    fixture_resources_for_ids,
    fixture_tags_xml,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

TAG1_ID1_ID2 = fixture_tags_xml([("tag1", ("id1", "id2"))])


class TestTagCreate(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            resources=fixture_resources_for_ids(),
        )

    def test_success_create(self):
        self.config.env.push_cib(tags=TAG1_ID1_ID2)
        cmd_tag.create(self.env_assist.get_env(), "tag1", ["id1", "id2"])

    def test_invalid_tag_id(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.create(
                self.env_assist.get_env(),
                "#tag",
                ["id1", "id2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_invalid_id("#tag", "#"),
            ]
        )

    def test_multiple_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.create(
                self.env_assist.get_env(),
                "",
                ["", ""],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_ID_IS_EMPTY,
                    id_description="id",
                ),
                fixture.error(reports.codes.TAG_CANNOT_CONTAIN_ITSELF),
                *[
                    fixture.report_not_found(_id, context_type="resources")
                    for _id in ["", ""]
                ],
                fixture.error(
                    reports.codes.TAG_ADD_REMOVE_IDS_DUPLICATION,
                    duplicate_ids_list=[""],
                    add_or_not_remove=True,
                ),
            ]
        )



from pcs import usage
from pcs.cli import dr
from pcs.cli.common.routing import create_router

dr_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.dr(argv)),
        "config": dr.config,
        "destroy": dr.destroy,
        "set-recovery-site": dr.set_recovery_site,
        "status": dr.status,
    },
    ["dr"],
)



import sys

from pcs import settings


def add_bundled_packages_to_path() -> None:
    """
    This function deals with some bundled python dependencies that are
    installed in a pcs-specific location rather than in a standard system
    location for the python packages.
    """
    if settings.pcs_bundled_packages_dir not in sys.path:
        sys.path.insert(0, settings.pcs_bundled_packages_dir)



# pylint: disable=global-statement
import dataclasses
import inspect
import multiprocessing as mp
import os
import signal
from logging import (
    Logger,
    getLogger,
)
from typing import (
    Any,
    Tuple,
    Union,
)

import dacite

from pcs.common import reports
from pcs.common.async_tasks.dto import CommandOptionsDto
from pcs.common.async_tasks.types import TaskFinishType
from pcs.common.interface import dto
from pcs.lib.auth.tools import (
    DesiredUser,
    get_effective_user,
)
from pcs.lib.auth.types import AuthUser
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.permissions.checker import PermissionsChecker
from pcs.utils import read_known_hosts_file_not_cached

from .command_mapping import (
    API_V1_COMPATIBILITY_MODE,
    COMMAND_MAP,
)
from .communicator import WorkerCommunicator
from .logging import (
    WORKER_LOGGER,
    setup_worker_logger,
)
from .report_processor import WorkerReportProcessor
from .types import (
    Message,
    TaskExecuted,
    TaskFinished,
    WorkerCommand,
)

worker_com: WorkerCommunicator


def _sigterm_handler(sig_num: int, frame: Any) -> None:
    del sig_num, frame
    if worker_com.is_locked:
        worker_com.set_terminate()
    else:
        raise SystemExit(0)


def worker_init(message_q: mp.Queue, logging_q: mp.Queue) -> None:
    """
    Runs in every new worker process after its creation
    :param message_q: Queue instance for sending messages to the scheduler
    :param logging_q: Queue instance for sending log records to the scheduler
    """
    # Create and configure new logger
    logger = setup_worker_logger(logging_q)
    logger.info("Worker initialized.")

    # Let task_executor use worker_com for sending messages to the scheduler
    global worker_com
    worker_com = WorkerCommunicator(message_q)

    def ignore_signals(sig_num, frame):  # type: ignore
        # pylint: disable=unused-argument
        pass

    signal.signal(signal.SIGINT, ignore_signals)
    signal.signal(signal.SIGTERM, _sigterm_handler)


def _pause_worker() -> None:
    logger = getLogger(WORKER_LOGGER)
    logger.debug(
        "Pausing worker until the scheduler updates status of this task."
    )
    os.kill(os.getpid(), signal.SIGSTOP)
    logger.debug("Worker unpaused.")


def _get_effective_user(
    logger: Logger, real_user: AuthUser, options: CommandOptionsDto
) -> AuthUser:
    effective_user = get_effective_user(
        real_user,
        DesiredUser(options.effective_username, options.effective_groups),
    )
    if real_user != effective_user:
        logger.debug(
            "Effective user=%s groups=%s",
            effective_user.username,
            ",".join(effective_user.groups),
        )
    return effective_user


def task_executor(task: WorkerCommand) -> None:
    """
    Launches the task inside the worker
    :param task: Task identifier, command and parameter object
    """
    logger = getLogger(WORKER_LOGGER)

    worker_com.put(
        Message(
            task.task_ident,
            TaskExecuted(os.getpid()),
        )
    )
    logger.info(
        "Task '%s' executed by user '%s'.",
        task.task_ident,
        task.auth_user.username,
    )
    request_timeout = task.command.command_dto.options.request_timeout
    if request_timeout is not None and request_timeout <= 0:
        logger.warning(
            "Invalid value '%s' for option 'request_timeout'", request_timeout
        )
        request_timeout = None
    auth_user = task.auth_user
    logger.debug(
        "Real user=%s groups=%s",
        auth_user.username,
        ",".join(auth_user.groups),
    )
    command_dto = task.command.command_dto
    if auth_user.is_superuser:
        auth_user = _get_effective_user(logger, auth_user, command_dto.options)

    env = LibraryEnvironment(  # type: ignore
        logger,
        WorkerReportProcessor(worker_com, task.task_ident),
        known_hosts_getter=read_known_hosts_file_not_cached,
        user_login=auth_user.username,
        user_groups=auth_user.groups,
        request_timeout=request_timeout,
    )

    task_retval = None
    command_name = command_dto.command_name
    try:
        if command_name not in COMMAND_MAP or (
            not task.command.api_v1_compatible
            and command_name in API_V1_COMPATIBILITY_MODE
        ):
            raise LibraryError(
                reports.ReportItem.error(
                    reports.messages.CommandUnknown(command_name)
                )
            )
        cmd = COMMAND_MAP[command_name]
        if not PermissionsChecker(logger).is_authorized(
            auth_user, cmd.required_permission
        ):
            raise LibraryError(
                reports.ReportItem.error(reports.messages.NotAuthorized())
            )
        # Dacite will validate command.params against command signature.
        # Dacite works only with dataclasses so we need to dynamically create
        # one
        try:
            data = dto.from_dict(
                dataclasses.make_dataclass(
                    f"{command_name}_params",
                    [
                        _param_to_field_tuple(param)
                        for param in list(
                            inspect.signature(cmd.cmd).parameters.values()
                        )[1:]
                    ],
                ),
                command_dto.params,
                strict=True,
            ).__dict__  # type: ignore
        except dacite.DaciteError as e:
            # TODO: make custom message from exception without mentioning
            # dataclasses and fields
            raise LibraryError(
                reports.ReportItem.error(
                    reports.messages.CommandInvalidPayload(str(e))
                )
            ) from e
        except dto.PayloadConversionError as e:
            raise LibraryError(
                reports.ReportItem.error(
                    reports.messages.CommandInvalidPayload("")
                )
            ) from e

        task_retval = cmd.cmd(env, **data)
    except LibraryError as e:
        # Some code uses args for storing ReportList, sending them to the report
        # processor here

        for report in e.args:
            # pylint: disable=no-member
            worker_com.put(Message(task.task_ident, report.to_dto()))
        worker_com.put(
            Message(
                task.task_ident,
                TaskFinished(TaskFinishType.FAIL, None),
            )
        )
        logger.exception("Task %s raised a LibraryError.", task.task_ident)
        _pause_worker()
        return
    except Exception as e:  # pylint: disable=broad-except
        # For unhandled exceptions during execution
        worker_com.put(
            Message(
                task.task_ident,
                TaskFinished(TaskFinishType.UNHANDLED_EXCEPTION, None),
            )
        )
        logger.exception(
            "Task %s raised an unhandled exception: %s", task.task_ident, e
        )
        _pause_worker()
        return
    worker_com.put(
        Message(
            task.task_ident,
            TaskFinished(TaskFinishType.SUCCESS, task_retval),
        )
    )
    logger.info("Task %s finished.", task.task_ident)
    _pause_worker()


def _param_to_field_tuple(
    param: inspect.Parameter,
) -> Union[Tuple[str, Any], Tuple[str, Any, dataclasses.Field]]:
    field_type = Any
    if param.annotation != inspect.Parameter.empty:
        field_type = param.annotation
    if param.default != inspect.Parameter.empty:
        return (
            param.name,
            field_type,
            dataclasses.field(default=param.default),
        )
    return (param.name, field_type)



import dataclasses
from textwrap import dedent
from unittest import TestCase

from pcs.common.str_tools import indent
from pcs.lib.cib import rule
from pcs.lib.cib.rule.expression_part import BoolExpr


def _parsed_to_str(parsed):
    if isinstance(parsed, BoolExpr):
        str_args = []
        for arg in parsed.children:
            str_args.extend(_parsed_to_str(arg).splitlines())
        return "\n".join(
            [f"{parsed.__class__.__name__} {parsed.operator}"]
            + indent(str_args)
        )

    parts = [parsed.__class__.__name__]
    for field in dataclasses.fields(parsed):
        value = getattr(parsed, field.name)
        if value is not None:
            if field.name in ("duration_parts", "date_parts"):
                parts.append(f"{field.name}=(\n ")
                parts.extend([f"{key}={val}" for key, val in value])
                parts.append("\n)")
            else:
                parts.append(f"{field.name}={value}")
    return " ".join(parts)


class Parser(TestCase):
    def _assert_success(self, rule_string, rule_tree):
        self.assertEqual(
            rule_tree,
            _parsed_to_str(rule.parse_rule(rule_string)),
        )

    def test_success_trivial(self):
        test_data = [
            ("", "BoolExpr AND"),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_resource_op_simple(self):
        test_data = [
            (
                "resource ::",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr"""
                ),
            ),
            (
                "resource ::dummy",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr type=dummy"""
                ),
            ),
            (
                "resource ocf::",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr standard=ocf"""
                ),
            ),
            (
                "resource :pacemaker:",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr provider=pacemaker"""
                ),
            ),
            (
                "resource systemd::Dummy",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr standard=systemd type=Dummy"""
                ),
            ),
            (
                "resource ocf:pacemaker:",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr standard=ocf provider=pacemaker"""
                ),
            ),
            (
                "resource :pacemaker:Dummy",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr provider=pacemaker type=Dummy"""
                ),
            ),
            (
                "resource ocf:pacemaker:Dummy",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr standard=ocf provider=pacemaker type=Dummy"""
                ),
            ),
            (
                "op monitor",
                dedent(
                    """\
                    BoolExpr AND
                      OpExpr name=monitor"""
                ),
            ),
            (
                "op monitor interval=10",
                dedent(
                    """\
                    BoolExpr AND
                      OpExpr name=monitor interval=10"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_resource_op_complex(self):
        test_data = [
            (
                "resource ::dummy and op monitor",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr type=dummy
                      OpExpr name=monitor"""
                ),
            ),
            (
                "resource ::dummy or op monitor interval=15s",
                dedent(
                    """\
                    BoolExpr OR
                      RscExpr type=dummy
                      OpExpr name=monitor interval=15s"""
                ),
            ),
            (
                "op monitor and resource ::dummy",
                dedent(
                    """\
                    BoolExpr AND
                      OpExpr name=monitor
                      RscExpr type=dummy"""
                ),
            ),
            (
                "op monitor interval=5min or resource ::dummy",
                dedent(
                    """\
                    BoolExpr OR
                      OpExpr name=monitor interval=5min
                      RscExpr type=dummy"""
                ),
            ),
            (
                "(resource ::dummy or resource ::delay) and op monitor",
                dedent(
                    """\
                    BoolExpr AND
                      BoolExpr OR
                        RscExpr type=dummy
                        RscExpr type=delay
                      OpExpr name=monitor"""
                ),
            ),
            (
                "(op start and op stop) or resource ::dummy",
                dedent(
                    """\
                    BoolExpr OR
                      BoolExpr AND
                        OpExpr name=start
                        OpExpr name=stop
                      RscExpr type=dummy"""
                ),
            ),
            (
                "op monitor or (resource ::dummy and resource ::delay)",
                dedent(
                    """\
                    BoolExpr OR
                      OpExpr name=monitor
                      BoolExpr AND
                        RscExpr type=dummy
                        RscExpr type=delay"""
                ),
            ),
            (
                "resource ::dummy and (op start or op stop)",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr type=dummy
                      BoolExpr OR
                        OpExpr name=start
                        OpExpr name=stop"""
                ),
            ),
            (
                "resource ::dummy and resource ::delay and op monitor",
                dedent(
                    """\
                    BoolExpr AND
                      RscExpr type=dummy
                      RscExpr type=delay
                      OpExpr name=monitor"""
                ),
            ),
            (
                "resource ::rA or resource ::rB or resource ::rC and op monitor",
                dedent(
                    """\
                    BoolExpr AND
                      BoolExpr OR
                        RscExpr type=rA
                        RscExpr type=rB
                        RscExpr type=rC
                      OpExpr name=monitor"""
                ),
            ),
            (
                "op start and op stop and op monitor or resource ::delay",
                dedent(
                    """\
                    BoolExpr OR
                      BoolExpr AND
                        OpExpr name=start
                        OpExpr name=stop
                        OpExpr name=monitor
                      RscExpr type=delay"""
                ),
            ),
            (
                "(resource ::rA or resource ::rB or resource ::rC) and (op oX or op oY or op oZ)",
                dedent(
                    """\
                    BoolExpr AND
                      BoolExpr OR
                        RscExpr type=rA
                        RscExpr type=rB
                        RscExpr type=rC
                      BoolExpr OR
                        OpExpr name=oX
                        OpExpr name=oY
                        OpExpr name=oZ"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_node_attr_simple(self):
        test_data = [
            (
                "defined pingd",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=DEFINED attr_name=pingd"""
                ),
            ),
            (
                "not_defined pingd",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NOT_DEFINED attr_name=pingd"""
                ),
            ),
            (
                "#uname eq node1",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node1"""
                ),
            ),
            (
                "#uname ne node2",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node2"""
                ),
            ),
            (
                "int gt 123",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=GT attr_name=int attr_value=123"""
                ),
            ),
            (
                "int gte 123",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=GTE attr_name=int attr_value=123"""
                ),
            ),
            (
                "int lt 123",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=LT attr_name=int attr_value=123"""
                ),
            ),
            (
                "int lte 123",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=LTE attr_name=int attr_value=123"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_node_attr_with_optional_parts(self):
        test_data = [
            (
                "#uname eq string node1",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node1 attr_type=STRING"""
                ),
            ),
            (
                "#uname eq integer 12345",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=12345 attr_type=INTEGER"""
                ),
            ),
            (
                "#uname eq integer -12345",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=-12345 attr_type=INTEGER"""
                ),
            ),
            (
                "#uname eq number 12345",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=12345 attr_type=NUMBER"""
                ),
            ),
            (
                "#uname eq number -12345",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=-12345 attr_type=NUMBER"""
                ),
            ),
            (
                "#uname eq version 1",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=1 attr_type=VERSION"""
                ),
            ),
            (
                "#uname eq version 1.2.3",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=1.2.3 attr_type=VERSION"""
                ),
            ),
            (
                "#uname eq string string",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=string attr_type=STRING"""
                ),
            ),
            (
                "#uname eq string and",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=and attr_type=STRING"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_and_or(self):
        test_data = [
            (
                "#uname ne node1 and #uname ne node2",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node2"""
                ),
            ),
            (
                "#uname eq node1 or #uname eq node2",
                dedent(
                    """\
                    BoolExpr OR
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node1
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node2"""
                ),
            ),
            (
                "#uname ne node1 and #uname ne node2 and #uname ne node3",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node2
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node3"""
                ),
            ),
            (
                "#uname ne node1 and #uname ne node2 or #uname eq node3",
                dedent(
                    """\
                    BoolExpr OR
                      BoolExpr AND
                        NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                        NodeAttrExpr operator=NE attr_name=#uname attr_value=node2
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node3"""
                ),
            ),
            (
                "#uname eq node1 or #uname eq node2 and #uname ne node3",
                dedent(
                    """\
                    BoolExpr AND
                      BoolExpr OR
                        NodeAttrExpr operator=EQ attr_name=#uname attr_value=node1
                        NodeAttrExpr operator=EQ attr_name=#uname attr_value=node2
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node3"""
                ),
            ),
            (
                "defined pingd and pingd lte 1",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=DEFINED attr_name=pingd
                      NodeAttrExpr operator=LTE attr_name=pingd attr_value=1"""
                ),
            ),
            (
                "pingd gt 1 or not_defined pingd",
                dedent(
                    """\
                    BoolExpr OR
                      NodeAttrExpr operator=GT attr_name=pingd attr_value=1
                      NodeAttrExpr operator=NOT_DEFINED attr_name=pingd"""
                ),
            ),
            (
                "#uname ne string integer and #uname ne string version",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=integer attr_type=STRING
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=version attr_type=STRING"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_date_simple(self):
        test_data = [
            (
                "date gt 2014-06-26",
                dedent(
                    """\
                    BoolExpr AND
                      DateUnaryExpr operator=GT date=2014-06-26"""
                ),
            ),
            (
                "date lt 2014-06-26",
                dedent(
                    """\
                    BoolExpr AND
                      DateUnaryExpr operator=LT date=2014-06-26"""
                ),
            ),
            (
                "date in_range 2014-06-26 to 2014-07-26",
                dedent(
                    """\
                    BoolExpr AND
                      DateInRangeExpr date_start=2014-06-26 date_end=2014-07-26"""
                ),
            ),
            (
                "date in_range to 2014-07-26",
                dedent(
                    """\
                    BoolExpr AND
                      DateInRangeExpr date_end=2014-07-26"""
                ),
            ),
            (
                "date in_range 2014-06-26 to duration years=1",
                dedent(
                    """\
                    BoolExpr AND
                      DateInRangeExpr date_start=2014-06-26 duration_parts=(
                        years=1 
                      )"""
                ),
            ),
            (
                "date in_range 2014-06-26 to duration a=1 b=2 a=3",
                dedent(
                    """\
                    BoolExpr AND
                      DateInRangeExpr date_start=2014-06-26 duration_parts=(
                        a=1 b=2 a=3 
                      )"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_datespec(self):
        test_data = [
            (
                "date-spec hours=1",
                dedent(
                    """\
                    BoolExpr AND
                      DatespecExpr date_parts=(
                        hours=1 
                      )"""
                ),
            ),
            (
                "date-spec hours=1-14 months=1 hours=14-20",
                dedent(
                    """\
                    BoolExpr AND
                      DatespecExpr date_parts=(
                        hours=1-14 months=1 hours=14-20 
                      )"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_and_or_date_datespec(self):
        test_data = [
            (
                "#uname ne node1 and date-spec hours=1-12",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      DatespecExpr date_parts=(
                        hours=1-12 
                      )"""
                ),
            ),
            (
                "date-spec monthdays=1-12 or #uname ne node1",
                dedent(
                    """\
                    BoolExpr OR
                      DatespecExpr date_parts=(
                        monthdays=1-12 
                      )
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1"""
                ),
            ),
            (
                "date-spec monthdays=1-10 or date-spec monthdays=11-20",
                dedent(
                    """\
                    BoolExpr OR
                      DatespecExpr date_parts=(
                        monthdays=1-10 
                      )
                      DatespecExpr date_parts=(
                        monthdays=11-20 
                      )"""
                ),
            ),
            (
                "#uname ne node1 and date in_range 2014-06-26 to 2014-07-26",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      DateInRangeExpr date_start=2014-06-26 date_end=2014-07-26"""
                ),
            ),
            (
                "date in_range 2014-06-26 to 2014-07-26 and #uname ne node1",
                dedent(
                    """\
                    BoolExpr AND
                      DateInRangeExpr date_start=2014-06-26 date_end=2014-07-26
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_braces(self):
        test_data = [
            (
                "(date-spec hours=1)",
                dedent(
                    """\
                    BoolExpr AND
                      DatespecExpr date_parts=(
                        hours=1 
                      )"""
                ),
            ),
            (
                "(#uname eq node1)",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node1"""
                ),
            ),
            (
                "(defined pingd)",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=DEFINED attr_name=pingd"""
                ),
            ),
            (
                "(#uname ne node1 and #uname ne node2)",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node2"""
                ),
            ),
            (
                "(#uname ne node1) and (#uname ne node2)",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node2"""
                ),
            ),
            (
                "(#uname ne node1 and #uname ne node2) or #uname eq node3",
                dedent(
                    """\
                    BoolExpr OR
                      BoolExpr AND
                        NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                        NodeAttrExpr operator=NE attr_name=#uname attr_value=node2
                      NodeAttrExpr operator=EQ attr_name=#uname attr_value=node3"""
                ),
            ),
            (
                "#uname ne node1 and (#uname ne node2 or #uname eq node3)",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      BoolExpr OR
                        NodeAttrExpr operator=NE attr_name=#uname attr_value=node2
                        NodeAttrExpr operator=EQ attr_name=#uname attr_value=node3"""
                ),
            ),
            (
                "(((#uname ne node1) and (((#uname ne node2) or (#uname eq node3)))))",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=NE attr_name=#uname attr_value=node1
                      BoolExpr OR
                        NodeAttrExpr operator=NE attr_name=#uname attr_value=node2
                        NodeAttrExpr operator=EQ attr_name=#uname attr_value=node3"""
                ),
            ),
            (
                "(date in_range 2014-06-26 to 2014-07-26)",
                dedent(
                    """\
                    BoolExpr AND
                      DateInRangeExpr date_start=2014-06-26 date_end=2014-07-26"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_success_not_date(self):
        test_data = [
            (
                "date eq 2014-06-26",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=EQ attr_name=date attr_value=2014-06-26"""
                ),
            ),
            (
                "date gt string 2014-06-26",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=GT attr_name=date attr_value=2014-06-26 attr_type=STRING"""
                ),
            ),
            (
                "date gt integer 12345",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=GT attr_name=date attr_value=12345 attr_type=INTEGER"""
                ),
            ),
            (
                "date gt version 1.2.3",
                dedent(
                    """\
                    BoolExpr AND
                      NodeAttrExpr operator=GT attr_name=date attr_value=1.2.3 attr_type=VERSION"""
                ),
            ),
        ]
        for rule_string, rule_tree in test_data:
            with self.subTest(rule_string=rule_string):
                self._assert_success(rule_string, rule_tree)

    def test_not_valid_rule(self):
        test_data = [
            # node attr misc
            ("#uname", (1, 7, 6, "Expected 'eq'")),
            ("string node1", (1, 8, 7, "Expected 'eq'")),
            # node attr unary
            ("defined", (1, 8, 7, "Expected <attribute name>")),
            ("not_defined", (1, 12, 11, "Expected <attribute name>")),
            ("defined string pingd", (1, 16, 15, "Expected end of text")),
            ("defined date-spec hours=1", (1, 19, 18, "Expected end of text")),
            ("defined duration hours=1", (1, 18, 17, "Expected end of text")),
            # node attr binary
            ("eq", (1, 3, 2, "Expected 'eq'")),
            ("#uname eq", (1, 10, 9, "Expected <attribute value>")),
            ("#uname node1", (1, 8, 7, "Expected 'eq'")),
            ("eq #uname", (1, 4, 3, "Expected 'eq'")),
            ("eq lt", (1, 6, 5, "Expected <attribute value>")),
            ("string #uname eq node1", (1, 8, 7, "Expected 'eq'")),
            ("date-spec hours=1 eq node1", (1, 19, 18, "Expected end of text")),
            (
                "#uname eq date-spec hours=1",
                (1, 21, 20, "Expected end of text"),
            ),
            ("duration hours=1 eq node1", (1, 10, 9, "Expected 'eq'")),
            ("#uname eq duration hours=1", (1, 20, 19, "Expected end of text")),
            # node attr binary with optional parts
            ("string", (1, 7, 6, "Expected 'eq'")),
            ("#uname eq string", (1, 17, 16, "Expected <attribute value>")),
            ("string #uname eq node1", (1, 8, 7, "Expected 'eq'")),
            # resource, op
            ("resource", (1, 9, 8, "Expected 'eq'")),
            ("op", (1, 3, 2, "Expected 'eq'")),
            ("resource ::rA and", (1, 15, 14, "Expected end of text")),
            ("resource ::rA and op ", (1, 15, 14, "Expected end of text")),
            ("resource ::rA and (", (1, 15, 14, "Expected end of text")),
            # and, or
            ("and", (1, 4, 3, "Expected 'eq'")),
            ("or", (1, 3, 2, "Expected 'eq'")),
            ("#uname and node1", (1, 8, 7, "Expected 'eq'")),
            ("#uname or node1", (1, 8, 7, "Expected 'eq'")),
            ("#uname or eq", (1, 8, 7, "Expected 'eq'")),
            ("#uname eq node1 and node2", (1, 17, 16, "Expected end of text")),
            ("#uname eq node1 and", (1, 17, 16, "Expected end of text")),
            (
                "#uname eq node1 and #uname eq",
                (1, 17, 16, "Expected end of text"),
            ),
            ("and #uname eq node1", (1, 5, 4, "Expected 'eq'")),
            (
                "#uname ne node1 and duration hours=1",
                (1, 17, 16, "Expected end of text"),
            ),
            (
                "duration monthdays=1 or #uname ne node1",
                (1, 10, 9, "Expected 'eq'"),
            ),
            # date
            ("date in_range", (1, 14, 13, "Expected 'to'")),
            ("date in_range 2014-06-26", (1, 15, 14, "Expected 'to'")),
            ("date in_range 2014-06-26 to", (1, 28, 27, "Expected <date>")),
            ("in_range 2014-06-26 to 2014-07-26", (1, 10, 9, "Expected 'eq'")),
            (
                "date in_range #uname eq node1 to 2014-07-26",
                (1, 15, 14, "Expected 'to'"),
            ),
            (
                "date in_range 2014-06-26 to #uname eq node1",
                (1, 36, 35, "Expected end of text"),
            ),
            (
                "date in_range defined pingd to 2014-07-26",
                (1, 15, 14, "Expected 'to'"),
            ),
            (
                "date in_range 2014-06-26 to defined pingd",
                (1, 37, 36, "Expected end of text"),
            ),
            (
                "string date in_range 2014-06-26 to 2014-07-26",
                (1, 8, 7, "Expected 'eq'"),
            ),
            (
                "date in_range string 2014-06-26 to 2014-07-26",
                (1, 15, 14, "Expected 'to'"),
            ),
            (
                "date in_range 2014-06-26 to string 2014-07-26",
                (1, 36, 35, "Expected end of text"),
            ),
            (
                "date in_range 2014-06-26 string to 2014-07-26",
                (1, 15, 14, "Expected 'to'"),
            ),
            (
                "#uname in_range 2014-06-26 to 2014-07-26",
                (1, 8, 7, "Expected 'eq'"),
            ),
            # braces
            ("(#uname)", (1, 8, 7, "Expected 'eq'")),
            ("(", (1, 2, 1, "Expected 'date'")),
            ("()", (1, 2, 1, "Expected 'date'")),
            ("(#uname", (1, 8, 7, "Expected 'eq'")),
            ("(#uname eq", (1, 11, 10, "Expected <attribute value>")),
            # pyparsing 2 uses double quotes, pyparsing 3 uses single quotes
            ("(#uname eq node1", (1, 17, 16, {"Expected ')'", 'Expected ")"'})),
        ]
        for rule_string, exception_data in test_data:
            with self.subTest(rule_string=rule_string):
                with self.assertRaises(rule.RuleParseError) as cm:
                    rule.parse_rule(rule_string)
                e = cm.exception
                if isinstance(exception_data[3], set):
                    self.assertIn(e.msg, exception_data[3])
                    self.assertEqual(
                        exception_data[0:3], (e.lineno, e.colno, e.pos)
                    )
                else:
                    self.assertEqual(
                        exception_data, (e.lineno, e.colno, e.pos, e.msg)
                    )
                self.assertEqual(rule_string, e.rule_string)



from unittest import TestCase

from pcs import settings
from pcs.common import host


def _dest(addr, port):
    return dict(addr=addr, port=port)


class PcsKnownHostFromKnownHost(TestCase):
    def setUp(self):
        self.name = "host name"
        self.token = "token"

    def test_from_known_host(self):
        known_host = host.PcsKnownHost.from_known_host_file_dict(
            self.name,
            dict(
                dest_list=[_dest("addr1", "port1"), _dest("addr2", "port2")],
                token=self.token,
            ),
        )
        self.assertEqual(self.name, known_host.name)
        self.assertEqual(self.token, known_host.token)
        self.assertEqual(
            [
                host.Destination("addr1", "port1"),
                host.Destination("addr2", "port2"),
            ],
            known_host.dest_list,
        )
        self.assertEqual(host.Destination("addr1", "port1"), known_host.dest)

    def test_missing_token(self):
        def known_host_call():
            return host.PcsKnownHost.from_known_host_file_dict(
                self.name,
                dict(
                    dest_list=[
                        _dest("addr1", "port1"),
                        _dest("addr2", "port2"),
                    ],
                ),
            )

        self.assertRaises(KeyError, known_host_call)

    def test_missing_port(self):
        def known_host_call():
            return host.PcsKnownHost.from_known_host_file_dict(
                self.name,
                dict(
                    dest_list=[_dest("addr1", "port1"), dict(addr="addr2")],
                    token=self.token,
                ),
            )

        self.assertRaises(KeyError, known_host_call)

    def test_missing_addr(self):
        def known_host_call():
            return host.PcsKnownHost.from_known_host_file_dict(
                self.name,
                dict(
                    dest_list=[_dest("addr1", "port1"), dict(port="port2")],
                    token=self.token,
                ),
            )

        self.assertRaises(KeyError, known_host_call)

    def test_no_dest_list(self):
        def known_host_call():
            return host.PcsKnownHost.from_known_host_file_dict(
                self.name, dict(token=self.token)
            )

        self.assertRaises(KeyError, known_host_call)

    def test_dest_list_empty(self):
        def known_host_call():
            return host.PcsKnownHost.from_known_host_file_dict(
                self.name, dict(dest_list=[], token=self.token)
            )

        self.assertRaises(KeyError, known_host_call)


class PcsKnownHost(TestCase):
    def setUp(self):
        self.name = "host_name"
        self.default_dest = host.Destination(
            self.name, settings.pcsd_default_port
        )

    def test_dest_list_none(self):
        self.assertEqual(
            self.default_dest, host.PcsKnownHost(self.name, None, None).dest
        )

    def test_dest_list_empty(self):
        self.assertEqual(
            self.default_dest, host.PcsKnownHost(self.name, None, []).dest
        )



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.lib_wrapper import Library


class LibraryWrapperTest(TestCase):
    def test_raises_for_bad_path(self):
        mock_middleware_factory = mock.MagicMock()
        lib = Library("env", mock_middleware_factory)
        self.assertRaises(Exception, lambda: lib.no_valid_library_part)

    @mock.patch("pcs.cli.common.lib_wrapper.constraint_order.create_with_set")
    @mock.patch("pcs.cli.common.lib_wrapper.cli_env_to_lib_env")
    def test_bind_to_library(self, mock_cli_env_to_lib_env, mock_order_set):
        # pylint: disable=no-self-use
        lib_env = mock.MagicMock()
        lib_env.is_cib_live = True
        lib_env.is_corosync_conf_live = True
        mock_cli_env_to_lib_env.return_value = lib_env

        def dummy_middleware(next_in_line, env, *args, **kwargs):
            return next_in_line(env, *args, **kwargs)

        mock_middleware_factory = mock.MagicMock()
        mock_middleware_factory.cib = dummy_middleware
        mock_middleware_factory.corosync_conf_existing = dummy_middleware
        mock_env = mock.MagicMock()
        Library(
            mock_env, mock_middleware_factory
        ).constraint_order.create_with_set("first", second="third")

        mock_order_set.assert_called_once_with(lib_env, "first", second="third")



import os.path
from typing import Optional

from pcs import settings
from pcs.common import file_type_codes as code
from pcs.common.file import FileMetadata


def _for_booth_config(filename: str) -> FileMetadata:
    return FileMetadata(
        # The filename is expected to be complete (i.e. booth.conf) and verified
        # (i.e. no slashes in it). The caller is responsible for doing both.
        file_type_code=code.BOOTH_CONFIG,
        path=os.path.join(settings.booth_config_dir, filename),
        owner_user_name="root",
        owner_group_name="root",
        permissions=0o644,
        is_binary=False,
    )


def _for_booth_key(filename: str) -> FileMetadata:
    return FileMetadata(
        # The filename is expected to be complete (i.e. booth.key) and verified
        # (i.e. no slashes in it). The caller is responsible for doing both.
        file_type_code=code.BOOTH_KEY,
        path=os.path.join(settings.booth_config_dir, filename),
        owner_user_name=settings.pacemaker_uname,
        owner_group_name=settings.pacemaker_gname,
        permissions=settings.booth_authkey_file_mode,
        is_binary=True,
    )


def _for_corosync_conf() -> FileMetadata:
    return FileMetadata(
        file_type_code=code.COROSYNC_CONF,
        path=settings.corosync_conf_file,
        owner_user_name="root",
        owner_group_name="root",
        permissions=0o644,
        is_binary=False,
    )


def _for_pacemaker_authkey() -> FileMetadata:
    return FileMetadata(
        file_type_code=code.PACEMAKER_AUTHKEY,
        path=settings.pacemaker_authkey_file,
        owner_user_name=settings.pacemaker_uname,
        owner_group_name=settings.pacemaker_gname,
        permissions=0o400,
        is_binary=True,
    )


def _for_pcs_dr_config() -> FileMetadata:
    return FileMetadata(
        file_type_code=code.PCS_DR_CONFIG,
        path=settings.pcsd_dr_config_location,
        owner_user_name="root",
        owner_group_name="root",
        permissions=0o600,
        is_binary=False,
    )


def _for_pcs_known_hosts() -> FileMetadata:
    return FileMetadata(
        file_type_code=code.PCS_KNOWN_HOSTS,
        path=settings.pcsd_known_hosts_location,
        owner_user_name="root",
        owner_group_name="root",
        permissions=0o600,
        is_binary=False,
    )


def _for_pcs_users_conf() -> FileMetadata:
    return FileMetadata(
        file_type_code=code.PCS_USERS_CONF,
        path=settings.pcsd_users_conf_location,
        owner_user_name="root",
        owner_group_name="root",
        permissions=0o600,
        is_binary=False,
    )


def _for_pcs_settings_conf() -> FileMetadata:
    return FileMetadata(
        file_type_code=code.PCS_SETTINGS_CONF,
        path=settings.pcsd_settings_conf_location,
        owner_user_name="root",
        owner_group_name="root",
        permissions=0o600,
        is_binary=False,
    )


def for_file_type(
    file_type_code: code.FileTypeCode, filename: Optional[str] = None
) -> FileMetadata:
    # pylint: disable=too-many-return-statements
    if file_type_code == code.BOOTH_CONFIG:
        if not filename:
            raise AssertionError("filename must be set")
        return _for_booth_config(filename)
    if file_type_code == code.BOOTH_KEY:
        if not filename:
            raise AssertionError("filename must be set")
        return _for_booth_key(filename)
    if file_type_code == code.COROSYNC_CONF:
        return _for_corosync_conf()
    if file_type_code == code.PACEMAKER_AUTHKEY:
        return _for_pacemaker_authkey()
    if file_type_code == code.PCS_DR_CONFIG:
        return _for_pcs_dr_config()
    if file_type_code == code.PCS_KNOWN_HOSTS:
        return _for_pcs_known_hosts()
    if file_type_code == code.PCS_USERS_CONF:
        return _for_pcs_users_conf()
    if file_type_code == code.PCS_SETTINGS_CONF:
        return _for_pcs_settings_conf()

    raise AssertionError("Unknown file_type_code")
